"""
Provides a class for creating a GIF from a NIfTI image which iterates through its slices based
on a specified view ('coronal', 'sagittal', or 'axial').

Classes defined in the module:
    * :class:`NiftiGifCreator`: This class creates a GIF from a NIfTI image by iterating through its slices based on
      a specified view ('coronal', 'sagittal', or 'axial'). The GIF is then written to the
      specified output directory with a filename prefix specified by the user.

TODO:
    * Add functionality to quickly generate slices of images.
    * Add flexibility in :class:`NiftiGifCreator` to accept a NumPy array or a file-path. This could also be added
      to the image slice class.

"""

import os
from typing import Iterable
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as mpl_animation
from ..utils import image_io as pet_pim

nifty_loader = pet_pim.safe_load_4dpet_nifti


class NiftiGifCreator:
    """
    This class is designed to create a GIF from a NIfTI image by iterating through its slices based on a specified view
    ('coronal', 'sagittal', or 'axial' (or 'transverse'). The GIF is then written to the specified output directory
    with a filename prefix specified by the user.
    
    .. important::
        
        Makes a GIF of the image in the space defined by the nifty file.

    Attributes:
        path_to_image (str): Path to the NIfTI image file.
        view (str): Specifies on which plane the NIfTI image is sliced.
        output_directory (str): Directory where the GIF should be written.
        prefix (str): Filename prefix for the GIF.
        fig (matplotlib.figure.Figure): Matplotlib figure object for the GIF.
        ax (matplotlib.axes.Axes): Matplotlib axes object for the GIF.
        im_kw (dict): Dictionary of arguments passed to Axes.imshow for the GIF.
        image (numpy.ndarray): 3D numpy array representing the NIfTI image.
        vmax (float): Maximum value for the plot's color limit.
        ani_image (matplotlib.image.AxesImage): An animated image rendered in the figure.
        cbar (matplotlib.colorbar): Colorbar associated with ani_image.

    Example:
        An example of how to initialize and use the NiftiGifCreator class:

        .. code-block:: python

            import petpal.image_visualization as pet_vis

            gif_creator = pet_vis.NiftiGifCreator(path_to_image="./input.nii.gz",
                                                  view="axial",
                                                  output_directory="./output",
                                                  output_filename_prefix="test")
            gif_creator.make_gif()
            gif_creator.write_gif()
    """
    def __init__(self,
                 path_to_image: str,
                 view: str,
                 output_directory: str,
                 output_filename_prefix: str = "",
                 fig_title: str = "Patlak-$K_i$ Parametric Image",
                 cbar_label: str = "$K_i$ (Infusion Rate)"):
        """
        Initialize the :class:`NiftiGifCreator`. Assigns instance variables and sets up figure for animation.

        Args:
            path_to_image (str): absolute path to the image file.
            view (str): desired view for the gif. Can be 'coronal', 'sagittal', 'axial'.
            output_directory (str): absolute path to the directory where gif is to be stored.
            output_filename_prefix (str, optional): a string to be prepended to the filename of the gif.
            fig_title (str, optional): Title of the figure where the images will be plotted.
            cbar_label (str, optional): Label for the colorbar of the figure.

        Raises:
            ValueError: If `view` is not 'coronal', 'sagittal' or 'axial'
        """
        
        self.view = view.lower()
        self._validate_view()
        
        self.path_to_image = os.path.abspath(path_to_image)
        self.output_directory = os.path.abspath(output_directory)
        self.prefix = output_filename_prefix
        
        self.image = nifty_loader(self.path_to_image).get_fdata()
        
        self.vmax = max(np.max(self.image), np.abs(np.min(self.image)))
        
        self.fig, self.ax = plt.subplots(1, 1, constrained_layout=True)
        self.ani = None
        self.cbar = None
        
        self.im_kw = {'origin'      : 'lower',
                     'cmap'         : 'bwr',
                     'vmin'         : -self.vmax / 3.,
                     'vmax'         : self.vmax / 3.,
                     'interpolation': 'none'}
        
        self.ani_image = self.make_first_frame()
        self.set_figure_title_and_labels(title=fig_title, cbar_label=cbar_label)
    
    def _validate_view(self):
        """
        Validate the value of ``view``. It must be one of ['coronal', 'sagittal', 'axial', 'x', 'y', 'z'].

        Raises:
            ValueError: If ``view`` is not 'coronal', 'sagittal', 'axial', 'x', 'y', 'z'
        """
        if self.view not in ['coronal', 'sagittal', 'axial', 'x', 'y', 'z', 'transverse']:
            raise ValueError("Invalid view. Please choose from 'coronal', 'sagittal', 'axial' (or 'transverse'), "
                             "'x', 'y', or 'z'.")
    
    def make_first_frame(self):
        """
        Makes the first frame of the animation by plotting an image on ``ax`` according to ``view``.

        Returns:
            im (matplotlib.image.AxesImage): Image plotted in the first frame.
            
        Side Effects:
            Modifies ``ani_image`` by creating the first frame of the GIF
        """
        if self.view in ['x', 'sagittal']:
            img = self.image[0, :, :].T
        elif self.view in ['y', 'coronal']:
            img = self.image[:, 0, :].T
        else:
            img = self.image[:, :, 0].T
        out_im = self.ax.imshow(img, **self.im_kw)
        
        return out_im
    
    def set_figure_title_and_labels(self, title: str, cbar_label: str):
        """
        Sets the title of the figure and labels of the color bar.

        Args:
            title (str): Title to be set for the figure.
            cbar_label (str): Label to be set for the color bar.
            
        Side Effects:
            - Modifies ``cbar`` with a new colorbar
            - Changes the title of ``fig``
            - Modifies the x and y axes of ``ani_image``
            
        """
        self.cbar = self.fig.colorbar(self.ani_image, ax=self.ax, shrink=1.0)
        self.cbar.set_label(cbar_label, rotation=270)
        self.fig.suptitle(title)
        self.ani_image.axes.get_xaxis().set_visible(False)
        self.ani_image.axes.get_yaxis().set_visible(False)
        
    def update_frame(self, i):
        """
        Update function for matplotlib.animation.FuncAnimation. Updates the plotted image on the current axes.

        Args:
            i (int): Index of the current frame

        Returns:
            Tuple containing ``im`` (matplotlib.image.AxesImage), which is updated to the i-th frame according to
            ``view``.
            
        Side Effects:
            Modifies ``ani_image`` by updating its frame data
            
        """
        if self.view in ['x', 'sagittal']:
            img = self.image[i, :, :].T
        elif self.view in ['y', 'coronal']:
            img = self.image[:, i, :].T
        else:
            img = self.image[:, :, i].T
        
        self.ani_image.set_data(img)
        
        return self.ani_image,
    
    def make_gif(self, frames: Iterable = None):
        """
        Makes the GIF using FuncAnimation from matplotlib.animation module.

        Args:
            frames (Iterable, optional): Iterable of frame indices to be included in the GIF. If not provided, all
                frames will be used.
            
        Side Effects:
            Modifies ``ani`` by creating a new :class:`matplotlib.animation.FuncAnimation` object
        """
        if frames is None:
            tot_dims = self.image.shape
            if self.view in ['x', 'sagittal']:
                num_frames = tot_dims[0]
            elif self.view in ['y', 'coronal']:
                num_frames = tot_dims[1]
            else:
                num_frames = tot_dims[2]
            frames = range(1, num_frames)
        
        self.ani = mpl_animation.FuncAnimation(fig=self.fig,
                                               func=self.update_frame,
                                               frames=frames,
                                               blit=True)
    
    def write_gif(self, fps: int = 15):
        """
        Writes the GIF to the output directory with filename ``{prefix}_view-{view}.gif``.
        
        Side Effects:
            - Creates a GIF file at the path specified by ``output_directory`` and ``prefix``
            - Closes the ``fig`` matplotlib figure
            
        """
        out_path = os.path.join(self.output_directory, f'{self.prefix}_view-{self.view}.gif')
        self.ani.save(f"{out_path}", fps=fps, writer='pillow', dpi=100)
        plt.close(self.fig)
        


"""
Module to handle timing information of PET scans.
"""
from typing import Self
from dataclasses import dataclass
import numpy as np

from .image_io import load_metadata_for_nifti_with_same_filename


@dataclass
class ScanTimingInfo:
    """
    A data structure to represent and streamline access to timing information for image scans.

    This class encapsulates details about a scan's timing, including:
    - Start and end times of each scan frame.
    - Duration and center times of the frames.
    - Decay values (if applicable).

    Additionally, the class provides properties for easy conversion of timing values to minutes
    if the times are given in seconds and exceed a threshold (assumed to be 200.0 seconds).

    Attributes:
        duration (np.ndarray[float]): Array of frame durations.
        end (np.ndarray[float]): Array of frame end times.
        start (np.ndarray[float]): Array of frame start times.
        center (np.ndarray[float]): Array of frame center times (midpoints).
        decay (np.ndarray[float]): Array of decay coefficients for the scan frames.

    Properties:
        duration_in_mins (np.ndarray[float]):
            Returns the frame durations converted to minutes if `end` is >= 200.0 seconds.
            Otherwise, returns the original durations.

        end_in_mins (np.ndarray[float]):
            Returns the frame end times converted to minutes if `end` is >= 200.0 seconds.
            Otherwise, returns the original end times.

        start_in_mins (np.ndarray[float]):
            Returns the frame start times converted to minutes if `end` is >= 200.0 seconds.
            Otherwise, returns the original start times.

        center_in_mins (np.ndarray[float]):
            Returns the frame center times converted to minutes if `end` is >= 200.0 seconds.
            Otherwise, returns the original center times.

    Examples:

        .. code-block:: python

            import numpy as np
            from petpal.utils.scan_timing import ScanTimingInfo

            # Explicitly setting the attributes
            ## Define scan timing information
            duration = np.array([60.0, 120.0, 180.0])  # seconds
            start = np.array([0.0, 60.0, 180.0])
            end = np.array([60.0, 180.0, 360.0])
            center = (start + end) / 2.0  # Calculate the midpoints
            decay = np.array([1.0, 0.9, 0.8])  # Example decay values

            ## Create an instance of ScanTimingInfo
            scan_timing_info = ScanTimingInfo(duration=duration, end=end, start=start, center=center, decay=decay)

            ## Access original timing information
            print(scan_timing_info.duration)  # [ 60. 120. 180.]
            print(scan_timing_info.center)    # [30.  120. 270.]

            ## Access timing as minutes (when times exceed 200.0 seconds)
            print(scan_timing_info.duration_in_mins)  # [ 60. 120. 180.] (Unchanged)
            print(scan_timing_info.center_in_mins)    # [30. 120. 270.] (Unchanged)

            ## Example when `end` is greater than 200.0:
            scan_timing_info.end = np.array([300.0, 400.0, 500.0])  # Update end times
            print(scan_timing_info.end_in_mins)  # [5. 6.66666667 8.33333333] (Converted to minutes)
            print(scan_timing_info.start_in_mins)  # [0. 1. 3.] (Converted to minutes)

            # Getting the object directly from a nifti image file
            # assuming the metadata shares the name
            scan_timing_info = ScanTimingInfo.from_nifti("/path/to/image.nii.gz")

    """
    duration: np.ndarray[float]
    start: np.ndarray[float]
    end: np.ndarray[float]
    center: np.ndarray[float]
    decay: np.ndarray[float]

    @property
    def duration_in_mins(self) -> np.ndarray[float]:
        """
        Returns the frame durations in minutes. Validates values by checking if the final frame
        value is greater than 200: if so, then assumes values are in seconds and divides by 60.
        """
        if self.end[-1] >= 200.0:
            return self.duration / 60.0
        return self.duration

    @property
    def end_in_mins(self) -> np.ndarray[float]:
        """
        Returns the frame time ends in minutes. Validates values by checking if the final frame
        value is greater than 200: if so, then assumes values are in seconds and divides by 60.
        """
        if self.end[-1] >= 200.0:
            return self.end / 60.0
        return self.end

    @property
    def start_in_mins(self) -> np.ndarray[float]:
        """
        Returns the frame time starts in minutes. Validates values by checking if the final frame
        value is greater than 200: if so, then assumes values are in seconds and divides by 60.
        """
        if self.end[-1] >= 200.0:
            return self.start / 60.0
        return self.start

    @property
    def center_in_mins(self) -> np.ndarray[float]:
        """
        Returns the frame reference times in minutes. Validates values by checking if the final
        frame value is greater than 200: if so, then assumes values are in seconds and divides by
        60.
        """
        if self.end[-1] >= 200.0:
            return self.center / 60.0
        return self.center


    @classmethod
    def from_metadata(cls, metadata_dict: dict) -> Self:
        r"""
        Extracts frame timing information and decay factors from a json metadata.
        Expects that the JSON metadata has ``FrameDuration`` and ``DecayFactor`` or
        ``DecayCorrectionFactor`` keys.

        .. important::
            This function tries to infer `FrameTimesEnd` and `FrameTimesStart` from the frame
            durations if those keys are not present in the metadata file. If the scan is broken,
            this might generate incorrect results.


        Args:
            metadata_dict (dict): The metadata dictionary, loaded into memory.

        Returns:
            :class:`ScanTimingInfo`: Frame timing information with the following elements:
                - duration (np.ndarray): Frame durations in seconds.
                - start (np.ndarray): Frame start times in seconds.
                - end (np.ndarray): Frame end times in seconds.
                - center (np.ndarray): Frame center times in seconds.
                - decay (np.ndarray): Decay factors for each frame.
        """
        frm_dur = np.asarray(metadata_dict['FrameDuration'], float)
        try:
            frm_starts = np.asarray(metadata_dict['FrameTimesStart'], float)
        except KeyError:
            frm_starts = np.cumsum(frm_dur)-frm_dur
        try:
            frm_ends = np.asarray(metadata_dict['FrameTimesEnd'], float)
        except KeyError:
            frm_ends = frm_starts+frm_dur
        try:
            decay = np.asarray(metadata_dict['DecayCorrectionFactor'], float)
        except KeyError:
            decay = np.asarray(metadata_dict['DecayFactor'], float)
        try:
            frm_centers = np.asarray(metadata_dict['FrameReferenceTime'], float)
        except KeyError:
            frm_centers = np.asarray(frm_starts + frm_dur / 2.0, float)

        return cls(duration=frm_dur,
                   start=frm_starts,
                   end=frm_ends,
                   center=frm_centers,
                   decay=decay)

    @classmethod
    def from_nifti(cls, image_path: str) -> Self:
        r"""
        Extracts frame timing information and decay factors from a NIfTI image metadata.
        Expects that the JSON metadata file has ``FrameDuration`` and ``DecayFactor`` or
        ``DecayCorrectionFactor`` keys.

        .. important::
            This function tries to infer `FrameTimesEnd` and `FrameTimesStart` from the frame 
            durations if those keys are not present in the metadata file. If the scan is broken,
            this might generate incorrect results.


        Args:
            image_path (str): Path to the NIfTI image file.

        Returns:
           :class:`ScanTimingInfo`: Frame timing information with the following elements:
                - duration (np.ndarray): Frame durations in seconds.
                - start (np.ndarray): Frame start times in seconds.
                - end (np.ndarray): Frame end times in seconds.
                - center (np.ndarray): Frame center times in seconds.
                - decay (np.ndarray): Decay factors for each frame.
        """
        _meta_data = load_metadata_for_nifti_with_same_filename(image_path=image_path)
        return cls.from_metadata(metadata_dict=_meta_data)

    @classmethod
    def from_start_end(cls,
                       frame_starts: np.ndarray,
                       frame_ends: np.ndarray,
                       decay_correction_factor: np.ndarray | None=None) -> Self:
        """Infer timing properties based on start and end time.
        
        Args:
            frame_starts (np.ndarray): Start time of each frame.
            frame_ends (np.ndarray): End time of each frame.
            decay_correction_factor (np.ndarray | None): Decay correction factor, which can be
                optionally provided based on the type of analysis being done. If None, frame decay
                will be set to ones. Default None.

        Returns:
            scan_timing_info (ScanTimingInfo): ScanTimingInfo object with the correct start, end,
                duration, midpoint, and (optionally) decay correction for each frame.

        Raises:
            ValueError: If frame_starts, frame_ends, and decay_correction_factor (if provided) are
                not of identical shape.

        """
        if frame_starts.shape != frame_ends.shape:
            raise ValueError("frame_ends must have the same shape as frame_starts")

        frame_duration = frame_ends - frame_starts
        frame_midpoint = frame_starts + frame_duration / 2
        frame_decay = np.ones_like(frame_starts)

        if decay_correction_factor is None:
            frame_decay = np.ones_like(frame_starts, dtype=float)
        else:
            frame_decay = np.asarray(decay_correction_factor, dtype=float)
            if frame_decay.shape != frame_starts.shape:
                raise ValueError("decay_correction_factor must have the same shape as frame_starts")

        return cls(duration=frame_duration,
                   start=frame_starts,
                   end=frame_ends,
                   center=frame_midpoint,
                   decay=frame_decay)


def get_window_index_pairs_from_durations(frame_durations: np.ndarray,
                                          window_duration: float) -> np.ndarray:
    r"""
    Computes start and end index pairs for windows of a given size based on frame durations.

    Args:
        frame_durations (np.ndarray): Array of frame durations in seconds.
        window_duration (float): Window size in seconds.

    Returns:
        np.ndarray: Array of shape (2, N), where the first row contains start indices,
            and the second row contains end indices for each window.

    Raises:
        ValueError: If `window_duration` is less than or equal to 0.
        ValueError: If `window_duration` is greater than the total duration of all frames.
    """
    if window_duration <= 0:
        raise ValueError("Window duration has to be > 0")
    if window_duration > np.sum(frame_durations):
        raise ValueError("Window duration is longer than the whole scan.")
    window_edge_indices = [0]
    window_duration_rolling_sum = 0
    for frame_index, frame_dur in enumerate(frame_durations):
        window_duration_rolling_sum += frame_dur
        if window_duration_rolling_sum >= window_duration:
            window_edge_indices.append(frame_index + 1)
            window_duration_rolling_sum = 0
    if window_edge_indices[-1]!=len(frame_durations):
        window_edge_indices.append(len(frame_durations))
    window_start_indices = np.asarray(window_edge_indices[:-1])
    window_end_indices = np.asarray(window_edge_indices[1:])
    index_pairs = np.vstack((window_start_indices, window_end_indices))
    return index_pairs


def get_window_index_pairs_for_image(image_path: str, window_duration: float):
    """
    Computes start and end index pairs for windows of a given size
    based on the frame durations of a NIfTI image.

    Args:
        image_path (str): Path to the NIfTI image file.
        window_duration (float): Window size in seconds.

    Returns:
        np.ndarray: Array of shape (2, N), where the first row contains start indices,
            and the second row contains end indices for each window.

    Raises:
        ValueError: If `window_duration` is less than or equal to 0.
        ValueError: If `window_duration` is greater than the total duration of all frames.

    See Also:
        :func:`get_window_index_pairs_from_durations`
    """
    image_frame_info = ScanTimingInfo.from_nifti(image_path=image_path)
    return get_window_index_pairs_from_durations(frame_durations=image_frame_info.duration,
                                                 window_duration=window_duration)


def calculate_frame_reference_time(frame_duration: np.ndarray,
                                   frame_start: np.ndarray,
                                   half_life: float) -> np.ndarray:
    r"""Compute frame reference time as the time at which the average activity occurs.
    
    Equation comes from the `DICOM standard documentation
    <https://dicom.innolitics.com/ciods/positron-emission-tomography-image/pet-image/00541300>`_

    :math:`T_{ave}=\frac{1}{\lambda}ln\frac{\lambda T}{1-e^{-\lambda T}}`

    where lambda is the decay constant, :math:`\frac{ln2}{T_{1/2}}`, :math:`T_{1/2}` is the half
    life, and :math:`T` is the frame duration.

    Args:
        frame_duration (np.ndarray): Duration of each frame in seconds.
        frame_start (np.ndarray): Start time of each frame relative to scan start, in seconds.
        half_life (float): Radionuclide half life in seconds.

    Returns: 
        np.ndarray: Frame reference time for each frame in the scan in seconds.
    """
    decay_constant = np.log(2)/half_life
    decay_over_frame = decay_constant*frame_duration
    reference_time_delay = np.log((decay_over_frame)/(1-np.exp(-decay_over_frame)))/decay_constant
    frame_reference_time = frame_start + reference_time_delay
    return frame_reference_time



"""Tools for calculating TAC uncertainty for application to kinetic models."""
import numpy as np

from ..utils.time_activity_curve import TimeActivityCurve


class TacUncertainty:
    """Set uncertainty for kinetic modeling. Includes options for constant,
    calculated, or preset uncertainty.    
    """
    def __init__(self,
                 time_activity_curve: TimeActivityCurve):
        """Initialize TacUncertainty with provided arguments.

        Args:
            time_activity_curve (TimeActivityCurve): The time activity curve used to determine
                uncertainties to pass onto the kinetic model.
        """
        self.time_activity_curve = time_activity_curve


    @property
    def constant_uncertainty(self) -> np.ndarray:
        """Get constant uncertainty for the TAC.
        """
        uncertainty = np.ones_like(self.time_activity_curve.activity)
        return uncertainty


    @property
    def tac_uncertainty(self) -> np.ndarray:
        """Get uncertainty stored in the TAC itself to pass on to the kinetic model.

        Raises:
            ValueError: If the supplied TAC uncertainty has any NaN values.
        
        Returns:
            uncertainty_from_tac (np.ndarray): Uncertainty stored in the supplied TAC.
        """
        uncertainty_from_tac = self.time_activity_curve.uncertainty
        nan_locations = np.isnan(uncertainty_from_tac)
        if np.any(nan_locations):
            raise ValueError("Supplied TAC has NaN values in stored uncertianty array. Use an "
                             "uncertainty method other than 'tac'.")
        return uncertainty_from_tac


    @property
    def calculated_uncertainty(self):
        """The calculated uncertainty for the TAC.

        Currently placeholder function.
        """
        raise NotImplementedError('Calculated uncertainty model for PET not yet implemented. Use '
                                  'constant uncertainty instead.')


    def __call__(self, uncertainty_method: str='constant') -> np.ndarray:
        """Get the tac uncertainty corresponding to the identified method.
        
        Args:
            uncertainty_method (str): Uncertainty type to apply to the tac. Default
                'constant'.

        Returns:
            uncertainty (np.ndarray): The uncertainty applied to each time frame in the TAC.

        Raises:
            ValueError: If `uncertainty_method` is not one of: 'constant', 'calculated', or 
                'tac'.
        """
        match uncertainty_method.lower():
            case 'constant':
                return self.constant_uncertainty
            case 'tac':
                return self.tac_uncertainty
            case 'calculated':
                return self.calculated_uncertainty
            case _:
                raise ValueError("uncertainty_method must be one of: 'constant','calculated', "
                                f"'tac'. Got {uncertainty_method}.")



"""
Library for math functions for use elsewhere.
"""
import numpy as np
from scipy.ndimage import gaussian_filter
import ants

def weighted_sum_computation(pet_series: ants.core.ANTsImage | np.ndarray,
                             frame_duration: np.ndarray,
                             half_life: float,
                             frame_start: np.ndarray,
                             decay_correction: np.ndarray):
    """
    Weighted sum of a PET image based on time and re-corrected for decay correction.

    Args:
        image_frame_duration (np.ndarray): Duration of each frame in pet series
        half_life (float): Half life of tracer radioisotope in seconds.
        pet_series (np.ndarray): 4D PET image series, as a data array.
        image_frame_start (np.ndarray): Start time of each frame in pet series,
            measured with respect to scan TimeZero.
        image_decay_correction (np.ndarray): Decay correction factor that scales
            each frame in the pet series. 

    Returns:
        image_weighted_sum (np.ndarray): 3D PET image computed by reversing decay correction
            on the PET image series, scaling each frame by the frame duration, then re-applying
            decay correction and scaling the image to the full duration.

    See Also:
        * :meth:`petpal.image_operations_4d.weighted_series_sum`: Function where this is implemented.

    """
    decay_constant = np.log(2.0) / half_life
    image_total_duration = np.sum(frame_duration)
    total_decay = decay_constant * image_total_duration
    total_decay /= 1.0 - np.exp(-1.0 * decay_constant * image_total_duration)
    total_decay /= np.exp(-1 * decay_constant * frame_start[0])
    
    pet_series_scaled = pet_series[:, :, :] * frame_duration / decay_correction
    pet_series_sum_scaled = pet_series_scaled.sum(axis=3)
    image_weighted_sum = pet_series_sum_scaled * total_decay / image_total_duration
    return image_weighted_sum


def weighted_sum_computation_over_index_window(pet_series: ants.core.ANTsImage | np.ndarray,
                                               window_start_id: int,
                                               window_end_id: int,
                                               half_life: float,
                                               frame_durations: np.ndarray,
                                               frame_starts: np.ndarray,
                                               decay_factors: np.ndarray):
    """
    Computes a weighted sum over a specified index window for a 4D PET image series. See
    :func:`weighted_sum_computation` for the computation.

    Args:
        pet_series (ants.core.ANTsImage | np.ndarray): PET image series as an ANTs image or a NumPy array.
        window_start_id (int): Start index of the window.
        window_end_id (int): End index of the window.
        half_life (float): Radioactive tracer's half-life in seconds.
        frame_durations (np.ndarray): Frame durations in seconds corresponding to the PET image series.
        frame_starts (np.ndarray): Frame start times in seconds corresponding to the PET image series.
        decay_factors (np.ndarray): Decay correction factors for the PET image series.

    Returns:
        np.ndarray: Weighted sum for the image series over the specified window.
    """
    window_image_series = pet_series[:, :, :, window_start_id:window_end_id]

    sub_frm_dur = frame_durations[window_start_id:window_end_id]
    sub_frm_st = frame_starts[window_start_id:window_end_id]
    sub_frm_decay = decay_factors[window_start_id:window_end_id]

    window_wss = weighted_sum_computation(frame_duration=sub_frm_dur,
                                          half_life=half_life,
                                          pet_series=window_image_series,
                                          frame_start=sub_frm_st,
                                          decay_correction=sub_frm_decay)

    return window_wss


def gauss_blur_computation(input_image: np.ndarray,
                           blur_size_mm: float,
                           input_zooms: list,
                           use_fwhm: bool):
    """
    Applies a Gaussian blur to an array image. Function intended to be a
    wrapper to be applied by other methods.
    """
    if use_fwhm:
        blur_size = blur_size_mm / (2*np.sqrt(2*np.log(2)))
    else:
        blur_size = blur_size_mm

    sigma_x = blur_size / input_zooms[0]
    sigma_y = blur_size / input_zooms[1]
    sigma_z = blur_size / input_zooms[2]

    blur_image = gaussian_filter(input=input_image,
                                 sigma=(sigma_x,sigma_y,sigma_z),
                                 axes=(0,1,2))
    return blur_image



from . import kinetic_modeling
from . import input_function
from . import preproc
from . import utils
from . import visualizations
from . import pipelines
from . import meta
from . import io

def main():
    print("PET-PAL (Positron Emission Tomography Processing and Analysis Library)")
    
if __name__ == "__main__":
    main()
    


"""
Class for doing RTM analysis
"""
import os
import warnings
from typing import Union
import json
import numpy as np
from .fit_tac_with_rtms import FitTACWithRTMs
from .graphical_analysis import (get_index_from_threshold,
                                 km_multifit_analysis_to_jsons,
                                 km_multifit_analysis_to_tsv)
from .reference_tissue_models import (calc_k2prime_from_mrtm_2003_fit,
                                      calc_k2prime_from_mrtm_original_fit,
                                      calc_bp_from_mrtm2_2003_fit,
                                      calc_bp_from_mrtm_original_fit,
                                      calc_bp_from_mrtm_2003_fit)
from ..utils.time_activity_curve import TimeActivityCurve, safe_load_tac
from ..utils.time_activity_curve import MultiTACAnalysisMixin


class RTMAnalysis:
    r"""
    A class designed to carry out various Reference Tissue Model (RTM) analyses on Time Activity
    Curves (TACs).

    This class eases the process of conducting RTM analysis on TACs. Paths to both reference and
    region-of-interest (ROI) TACs are taken as inputs at initialization. The class further provides
    multiple utility functions for initializing and running the RTM analysis, and also for
    validating the inputs based on the RTM method chosen.

    This class currently supports various RTM methods such as :'srtm', 'frtm', 'mrtm-original',
    'mrtm', and 'mrtm2'.

    Attributes:
        ref_tac_path (str): Absolute path for reference TAC
        roi_tac_path (str): Absolute path for ROI TAC
        output_directory (str): Absolute path for the output directory
        output_filename_prefix (str): Prefix for the output filename of the result
        method (str): RTM analysis method. Converts to lower case at initialization.
        analysis_props (dict): Analysis properties dictionary initialized with method-specific 
            property keys and default values.
        _has_analysis_been_run (bool): Flag representing if the RTM analysis has been run to ensure
            correct order of operations.

    Example:
        In the proceeding example, we assume that we have two tacs: a reference region tac, and a
        region of interest (ROI) tac named 'ref_tac.txt' and 'roi_tac.txt', respectively.
        Furthermore, we assume that both TACs are sampled at the same times, and are evenly sampled
        with respect to time.

        .. code-block:: python

            import numpy as np
            from petpal.kinetic_modeling.rtm_analysis as pet_rtms

            file_rtm = pet_rtms.RTMAnalysis(ref_tac_path="ref_tac.txt",
                                            roi_tac_path="roi_tac.txt",
                                            output_directory="./",
                                            output_filename_prefix='pre',
                                            method="mrtm")
            file_rtm.run_analysis(t_thresh_in_mins=40.0)
            file_rtm.save_analysis()


    See Also:
        * :class:`FitTACWithRTMs`: a class for analyzing TACs with RTMs.

    """
    def __init__(self,
                 ref_tac_path: str,
                 roi_tac_path: str,
                 output_directory: str,
                 output_filename_prefix: str,
                 method: str):
        r"""
        Initialize RTMAnalysis with provided arguments.

        The init method executes the following operations:
            1. It converts the provided analysis method to lower case for consistency in internal
                processing.
            2. It obtains the absolute paths for reference and ROI TAC files and the output
                directory, to ensure they are consistently accessible.
            3. It initializes the analysis properties dictionary using `init_analysis_props` method.
            4. It initializes the `_has_analysis_been_run` flag to False, to indicate that the RTM
                analysis has not yet been run.

        Args:
            ref_tac_path (str): Path to the file containing reference TAC.
            roi_tac_path (str): Path to the file containing ROI TAC.
            output_directory (str): Path to the directory where the output will be saved.
            output_filename_prefix (str): Prefix that will be used for the output filename.
            method (str): The RTM analysis method to be used. Could be one of 'srtm', 'frtm',
                'mrtm-original', 'mrtm' or 'mrtm2'.

        """
        self.ref_tac_path: str = os.path.abspath(ref_tac_path)
        self.roi_tac_path: str = os.path.abspath(roi_tac_path)
        self.output_directory: str = os.path.abspath(output_directory)
        self.output_filename_prefix: str = output_filename_prefix
        self.method = method.lower()
        self.analysis_props: dict = self.init_analysis_props(self.method)
        self._has_analysis_been_run: bool = False

    def init_analysis_props(self, method: str) -> dict:
        r"""
        Initializes the analysis properties dict based on the specified RTM analysis method.

        Args:
            method (str): RTM analysis method. Must be one of 'srtm', 'frtm', 'mrtm-original',
                'mrtm' or 'mrtm2'.

        Returns:
            dict: A dictionary containing method-specific property keys and default values.

        Raises:
            ValueError: If input `method` is not one of the supported RTM methods.
        """
        common_props = {'FilePathRTAC': self.ref_tac_path,
                        'FilePathTTAC': self.roi_tac_path,
                        'MethodName': method.upper(),
                        'k2Prime': None}
        if method.startswith("mrtm"):
            props = {
                'BP': None,
                'k2Prime': None,
                'ThresholdTime': None,
                'StartFrameTime': None,
                'EndFrameTime' : None,
                'NumberOfPointsFit': None,
                'RawFits': None,
                'SimulatedFits': None,
                **common_props
                }
        elif method.startswith("srtm") or method.startswith("frtm"):
            props = {
                'FitValues': None,
                'FitStdErr': None,
                **common_props
                }
        else:
            raise ValueError(f"Invalid method! Must be either 'srtm', 'frtm', 'srtm2', 'frtm2', "
                             f"'mrtm-original', 'mrtm' or 'mrtm2'. Got {method}.")
        return props

    def run_analysis(self,
                     bounds: Union[None, np.ndarray] = None,
                     t_thresh_in_mins: float = None,
                     k2_prime: float = None,
                     **tac_load_kwargs):
        r"""
        Runs the full RTM analysis process which involves validating inputs, calculation fits, and
        deducing fit properties.

        Specifically, it executes the following sequence:
            1. :meth:`validate_analysis_inputs`
            2. :meth:`calculate_fit`
            3. :meth:`calculate_fit_properties`

        Args:
            bounds (Union[None, np.ndarray], optional): Optional boundaries for parameters for
                fitting function.
            t_thresh_in_mins (float, optional): Threshold time in minutes for the MRTM analyses.
            k2_prime (float, optional): Input for the modified RTM (MRTM2, FRTM2, and SRTM2)
                analyses.

        Returns:
            None
        """
        self.validate_analysis_inputs(k2_prime=k2_prime, t_thresh_in_mins=t_thresh_in_mins)

        fit_results = self.calculate_fit(bounds=bounds,
                                         t_thresh_in_mins=t_thresh_in_mins,
                                         k2_prime=k2_prime,
                                         **tac_load_kwargs)
        self.calculate_fit_properties(fit_results=fit_results,
                                      t_thresh_in_mins=t_thresh_in_mins,
                                      k2_prime=k2_prime)
        self._has_analysis_been_run = True

    def validate_analysis_inputs(self, k2_prime, t_thresh_in_mins):
        r"""
        Validates the provided inputs for the RTM analysis.

        If MRTM type of analysis is being run, it ensures that ``t_thresh_in_mins`` is not None.
        If modified analysis is being done (MRTM2, FRTM2, SRTM2), it ensures ``k2_prime`` is not
        None.

        Args:
            k2_prime (float): k2 prime value.
            t_thresh_in_mins (float): Threshold time for MRTM analyses.

        Raises:
            ValueError: If an input required for the selected method is `None`.
        """
        if self.method.startswith("mrtm") and t_thresh_in_mins is None:
            raise ValueError("t_thresh_in_mins must be set for the MRTM analyses.")
        if self.method.endswith("2") and k2_prime is None:
            raise ValueError("k2_prime must be set for the modified RTM (MRTM2, FRTM2, and SRTM2) "
                             "analyses.")

    def calculate_fit(self,
                      bounds: Union[None, np.ndarray] = None,
                      t_thresh_in_mins: float = None,
                      k2_prime: float = None):
        r"""
        Calculates the model fitting parameters for TACs using the chosen RTM analysis method.

        This method executes the following sequence:
            1. :meth:`validate_analysis_inputs`
            2. :meth:`safe_load_tac` for both reference and ROI TACs
            3. Creates a :class:`FitTACWithRTMs` instance and fits TAC to the model

        Args:
            bounds (Union[None, np.ndarray]): Boundaries for parameters for fitting function.
            t_thresh_in_mins (float): Threshold time for MRTM analyses.
            k2_prime (float): k2 prime value.
            tac_load_kwargs (Any): Additional keyword arguments for the loading TAC function.

        Returns:
            FitResults: Object containing fit results.
        """
        self.validate_analysis_inputs(k2_prime=k2_prime, t_thresh_in_mins=t_thresh_in_mins)

        reference_tac = TimeActivityCurve.from_tsv(filename=self.ref_tac_path)
        target_tac = TimeActivityCurve.from_tsv(filename=self.roi_tac_path)
        analysis_obj = FitTACWithRTMs(target_tac=target_tac,
                                      reference_tac=reference_tac,
                                      method=self.method,
                                      bounds=bounds,
                                      t_thresh_in_mins=t_thresh_in_mins,
                                      k2_prime=k2_prime)
        analysis_obj.fit_tac_to_model()

        return analysis_obj.fit_results

    def calculate_fit_properties(self,
                                 fit_results: Union[np.ndarray, tuple[np.ndarray, np.ndarray]],
                                 t_thresh_in_mins: float = None,
                                 k2_prime: float = None):
        r"""
        Calculates additional fitting properties based on the raw fit results.

        It delegates the calculation to method-specific functions:
            1. For 'srtm' or 'frtm' methods: :meth:`_calc_frtm_or_srtm_fit_props` is used.
            2. For 'mrtm' methods: :meth:`_calc_mrtm_fit_props` is used.

        Args:
            fit_results (Union[np.ndarray, tuple[np.ndarray, np.ndarray]]): The fit results.
            t_thresh_in_mins (float): Threshold time for MRTM analyses.
            k2_prime (float): k2 prime value for 'mrtm' based methods.

        Returns:
            None
        """
        if self.method.startswith("frtm") or self.method.startswith("srtm"):
            self._calc_frtm_or_srtm_fit_props(fit_results=fit_results,
                                              k2_prime=k2_prime,
                                              props_dict=self.analysis_props)
        else:
            self._calc_mrtm_fit_props(fit_results=fit_results,
                                      k2_prime=k2_prime,
                                      t_thresh_in_mins=t_thresh_in_mins,
                                      props_dict=self.analysis_props)

    def save_analysis(self):
        r"""
        Save the analysis results in JSON format.

        The results are only saved if the analysis has been run (_has_analysis_been_run flag is
        checked).

        Raises:
            RuntimeError: If the :meth:'run_analysis' method has not been called yet.
        """
        if not self._has_analysis_been_run:
            raise RuntimeError("'run_analysis' method must be called before 'save_analysis'.")
        file_name_prefix = os.path.join(self.output_directory,
                                        f"{self.output_filename_prefix}_analysis-"
                                        f"{self.analysis_props['MethodName']}")
        analysis_props_file = f"{file_name_prefix}_props.json"
        with open(analysis_props_file, 'w', encoding='utf-8') as f:
            json.dump(obj=self.analysis_props, fp=f, indent=4)

    def __call__(self, bounds, t_thresh_in_mins, k2_prime):
        self.run_analysis(bounds=bounds, t_thresh_in_mins=t_thresh_in_mins, k2_prime=k2_prime)
        self.save_analysis()

    def _calc_mrtm_fit_props(self, fit_results: Union[np.ndarray, tuple[np.ndarray, np.ndarray]],
                             k2_prime: float,
                             t_thresh_in_mins: float,
                             props_dict: dict,
                             write_simulated: bool=False):
        r"""
        Internal function used to calculate additional fitting properties for 'mrtm' type analyses.

        This method is used internally within :meth:`calculate_fit_properties`.

        Args:
            fit_results (np.ndarray): Resulting fit parameters.
            k2_prime (float): k2 prime value for 'mrtm' based methods.
            t_thresh_in_mins (float): Threshold time for MRTM analyses.
        """
        self.validate_analysis_inputs(k2_prime=k2_prime, t_thresh_in_mins=t_thresh_in_mins)
        fit_ans, y_fit = fit_results
        if self.method == 'mrtm-original':
            bp_val = calc_bp_from_mrtm_original_fit(fit_ans)
            k2_val = calc_k2prime_from_mrtm_original_fit(fit_ans)
        elif self.method == 'mrtm':
            bp_val = calc_bp_from_mrtm_2003_fit(fit_ans)
            k2_val = calc_k2prime_from_mrtm_2003_fit(fit_ans)
        else:
            bp_val = calc_bp_from_mrtm2_2003_fit(fit_ans)
            k2_val = None
        props_dict["k2Prime"] = k2_val
        props_dict["BP"] = bp_val
        props_dict["RawFits"] = list(fit_ans)

        if write_simulated:
            props_dict["SimulatedFits"] = list(y_fit.round(7))

        ref_tac_times, _ = safe_load_tac(filename=self.ref_tac_path)
        t_thresh_index = get_index_from_threshold(times_in_minutes=ref_tac_times,
                                                  t_thresh_in_minutes=t_thresh_in_mins)
        props_dict['ThresholdTime'] = t_thresh_in_mins
        props_dict['StartFrameTime'] = ref_tac_times[t_thresh_index]
        props_dict['EndFrameTime'] = ref_tac_times[-1]
        props_dict['NumberOfPointsFit'] = len(ref_tac_times[t_thresh_index:])

    def _calc_frtm_or_srtm_fit_props(self,
                                     fit_results: tuple[np.ndarray, np.ndarray],
                                     k2_prime: float,
                                     props_dict: dict):
        r"""
        Internal function used to calculate additional fitting properties for 'frtm' and 'srtm'
        type analyses.

        This method is used internally within :meth:`calculate_fit_properties`.

        Args:
            fit_results (tuple[np.ndarray, np.ndarray]): Tuple containing the fit parameters and
                their corresponding fit covariances.

        """
        fit_params, fit_covariances = fit_results
        fit_stderr = np.sqrt(np.diagonal(fit_covariances))

        if self.method.startswith('srtm'):
            format_func =  self._get_pretty_srtm_fit_param_vals
        else:
            format_func = self._get_pretty_frtm_fit_param_vals

        if self.method.endswith('2'):
            props_dict["k2Prime"] = k2_prime
            props_dict["FitValues"] = format_func(fit_params.round(5), True)
            props_dict["FitStdErr"] = format_func(fit_stderr.round(5), True)
        else:
            props_dict["FitValues"] = format_func(fit_params.round(5), False)
            props_dict["FitStdErr"] = format_func(fit_stderr.round(5), False)

    @staticmethod
    def _get_pretty_srtm_fit_param_vals(param_fits: np.ndarray, reduced: bool = False) -> dict:
        r"""
        Utility function to get nicely formatted fit parameters for 'srtm(2)' analysis.

        Returns a dictionary with keys: 'R1', 'k2', and 'BP' and the corresponding values from
        ``param_fits``.

        Args:
            param_fits (np.ndarray): array containing the fit parameters.

        Returns:
            dict: Dictionary of fit parameters and their corresponding values.
        """
        if reduced:
            return {name: val for name, val in zip(['R1', 'BP'], param_fits)}
        else:
            return {name: val for name, val in zip(['R1', 'k2', 'BP'], param_fits)}

    @staticmethod
    def _get_pretty_frtm_fit_param_vals(param_fits: np.ndarray, reduced: bool = False) -> dict:
        r"""
        Utility function to get nicely formatted fit parameters for 'frtm(2)' analysis.

        Returns a dictionary with keys: 'R1', 'k2', 'k3', and 'k4' and the corresponding values from
        ``param_fits``.

        Args:
            param_fits (np.ndarray): array containing the fit parameters.

        Returns:
            dict: Dictionary of fit parameters and their corresponding values.
        """
        if reduced:
            return {name: val for name, val in zip(['R1', 'k3', 'k4'], param_fits)}
        else:
            return {name: val for name, val in zip(['R1', 'k2', 'k3', 'k4'], param_fits)}


class MultiTACRTMAnalysis(RTMAnalysis, MultiTACAnalysisMixin):
    """
    A class for performing reference tissue model (RTM) analysis on multiple tissue TACs.

    Attributes:
        ref_tac_path (str): Path to the reference TAC file.
        roi_tacs_dir (str): Directory containing region of interest TAC files.
        output_directory (str): Directory for saving analysis results.
        output_filename_prefix (str): Prefix for output filenames.
        method (str): Method used for RTM analysis.
    """
    def __init__(self,
                 ref_tac_path: str,
                 roi_tacs_dir: str,
                 output_directory: str,
                 output_filename_prefix: str,
                 method: str):
        """
        Initializes the MultiTACRTMAnalysis object with required paths and analysis method.

        Args:
            ref_tac_path (str): Path to the reference TAC file.
            roi_tacs_dir (str): Directory containing region of interest TAC files.
            output_directory (str): Directory for saving analysis results.
            output_filename_prefix (str): Prefix for output filenames.
            method (str): Method used for RTM analysis.
        """
        MultiTACAnalysisMixin.__init__(self,
                                       input_tac_path=ref_tac_path,
                                       tacs_dir=roi_tacs_dir,)
        RTMAnalysis.__init__(self,
                             ref_tac_path=ref_tac_path,
                             roi_tac_path=roi_tacs_dir,
                             output_directory=output_directory,
                             output_filename_prefix=output_filename_prefix,
                             method=method)


    def init_analysis_props(self, method: str) -> list[dict]:
        """
        Initializes analysis properties for each tissue TAC using the specified method.
        Overrides :meth:`RTMAnalysis.init_analysis_props`.

        Args:
            method (str): Method used for initializing analysis properties.

        Returns:
            list[dict]: A list of analysis property dictionaries for each TAC.
        """
        num_of_tacs = self.num_of_tacs
        analysis_props = [RTMAnalysis.init_analysis_props(self, method=method) for a_tac in range(num_of_tacs)]
        for tac_id, a_prop_dict in enumerate(analysis_props):
            a_prop_dict['FilePathTTAC'] = os.path.abspath(self.tacs_files_list[tac_id])
        return analysis_props


    def calculate_fit(self,
                      bounds: Union[None, np.ndarray] = None,
                      t_thresh_in_mins: float = None,
                      k2_prime: float = None) -> list:
        """
        Calculates the fit for each TAC, updating the analysis properties with model fit results.
        Overrides :meth:`RTMAnalysis.calculate_fit`.

        Args:
            bounds (Union[None, np.ndarray], optional): Bounds for the fitting parameters. Defaults to None.
            t_thresh_in_mins (float, optional): Threshold in minutes for fit calculation. Defaults to None.
            k2_prime (float, optional): A reference tissue model parameter. Defaults to None.
            **tac_load_kwargs: Additional keyword arguments for TAC loading.

        Returns:
            list: A list of fit results for each TAC.
        """
        reference_tac = TimeActivityCurve.from_tsv(filename=self.ref_tac_path)
        fit_results = []
        for _, a_tac in enumerate(self.tacs_files_list):
            target_tac = TimeActivityCurve.from_tsv(filename=a_tac)
            analysis_obj = FitTACWithRTMs(reference_tac=reference_tac,
                                          target_tac=target_tac,
                                          method=self.method,
                                          bounds=bounds,
                                          t_thresh_in_mins=t_thresh_in_mins,
                                          k2_prime=k2_prime)
            analysis_obj.fit_tac_to_model()
            fit_results.append(analysis_obj.fit_results)

        return fit_results

    def calculate_fit_properties(self,
                                 fit_results: list[np.ndarray],
                                 t_thresh_in_mins: float = None,
                                 k2_prime: float = None):
        """
        Calculates additional properties of the fit based on the fit results and threshold for each TAC.
        Overrides :meth:`RTMAnalysis.calculate_fit_properties`.

        Args:
            fit_results (list[np.ndarray]): Results from the fitting procedure.
            t_thresh_in_mins (float, optional): Threshold in minutes used for fitting. Defaults to None.
            k2_prime (float, optional): A reference tissue model parameter. Defaults to None.
        """
        if self.method.startswith("frtm") or self.method.startswith("srtm"):
            for props_dict, fit_vals in zip(self.analysis_props, fit_results):
                self._calc_frtm_or_srtm_fit_props(fit_results=fit_vals,
                                                  k2_prime=k2_prime,
                                                  props_dict=props_dict)
        else:
            for props_dict, fit_vals in zip(self.analysis_props, fit_results):
                self._calc_mrtm_fit_props(fit_results=fit_vals,
                                          k2_prime=k2_prime,
                                          t_thresh_in_mins=t_thresh_in_mins,
                                          props_dict=props_dict)

    def save_analysis(self, output_as_tsv: bool=True, output_as_json: bool=False):
        """
        Saves the analysis results to a TSV file as a table with fit parameters for each ROI.

        Args:
            output_as_tsv (bool): Set True to write results to TSV table. Default True.
            output_as_json (bool): Set True to write results to a folder with one JSON file per
                region. Default False.

        Raises:
            RuntimeError: If 'run_analysis' method has not been called before save_analysis.
        """
        if not self._has_analysis_been_run:
            raise RuntimeError("'run_analysis' method must be called before 'save_analysis'.")

        if output_as_tsv:
            km_multifit_analysis_to_jsons(analysis_props=self.analysis_props,
                                          output_directory=self.output_directory,
                                          output_filename_prefix=self.output_filename_prefix,
                                          method=self.method,
                                          inferred_seg_labels=self.inferred_seg_labels)
        if output_as_json:
            km_multifit_analysis_to_tsv(analysis_props=self.analysis_props,
                                        output_directory=self.output_directory,
                                        output_filename_prefix=self.output_filename_prefix,
                                        method=self.method,
                                        inferred_seg_labels=self.inferred_seg_labels)
        if not output_as_tsv and not output_as_json:
            warnings.warn('Both output_as_tsv and output_as_json set False. Results not written to '
                          'file.')

    def __call__(self, output_as_tsv: bool=True, output_as_json: bool=False, **run_kwargs):
        """
        Runs :meth:`run_analysis` and :meth:`save_analysis` to run the analysis and save the
        analysis properties.
        
        Args:
            output_as_tsv (bool): Set True to write results to TSV table. Default True.
            output_as_json (bool): Set True to write results to a folder with one JSON file per
                region. Default False.
            run_kwargs: Additional keyword arguments used in the analysis. These are passed on to
                :meth:`run_analysis`.
        """
        self.run_analysis(**run_kwargs)
        self.save_analysis(output_as_tsv=output_as_tsv, output_as_json=output_as_json)



"""
Module for loading and saving images
"""
from typing import Optional
from collections.abc import Callable
import ants


class ImageLoader:
    """Class for reading image files, with extensions such as Nifti or MGZ.

    See also: :py:docs:`~ants.image_read`.

    Example:

        .. code-block:: python

            from petpal.io.image import ImageLoader

            image_loader = ImageLoader()
            my_img = image_loader.load('/path/to/img.nii.gz')

    :ivar _loader: Function that loads an image file as an ants.ANTsImage object.
    """
    def __init__(self, loader: Optional[Callable[[str], ants.ANTsImage]] = None):
        self._loader = loader or ants.image_read

    def load(self, filename: str) -> ants.ANTsImage:
        """Public read API that delegates to the configured reader.
        
        Args:
            filename (str): Path to file that will be loaded as ANTsImage.
    
        Returns:
            img (ants.ANTsImage): Image object loaded into Python."""
        return self._loader(filename)

    def __call__(self, filename: str) -> ants.ANTsImage:
        return self.load(filename=filename)



"""CLI command for running a basic PiB pipeline on a subject, generating SUVR parametric image and tacs."""

import argparse
import petpal
from petpal.pipelines.preproc_steps import ImageToImageStep, TACsFromSegmentationStep
from petpal.pipelines.steps_containers import StepsContainer
from petpal.pipelines.pipelines import BIDS_Pipeline

_PIB_EXAMPLE_ = r"""
Example:
    - Run a PIB scan through SUVR pipeline:
      'petpal-pib-proc --sub 001 --ses 01 --pet-path ./sub-001/ses-01/pet/sub-001_ses-01_pet.nii.gz 
                --anat-path ./sub-001/ses-01/anat/sub-001_ses-01_T1w.nii.gz --seg-path 
                ./derivatives/freesurfer/sub-001/ses-01/mri/wmparc.mgz --seg-table-path 
                ./dseg.tsv --ref-region-label 8'
                 
    - Only compute SUVR for the range of 30-60 min:
      'petpal-pib-proc --sub 001 --ses 01 --pet-path ./sub-001/ses-01/pet/sub-001_ses-01_pet.nii.gz 
                --anat-path ./sub-001/ses-01/anat/sub-001_ses-01_T1w.nii.gz --seg-path 
                ./derivatives/freesurfer/sub-001/ses-01/mri/wmparc.mgz --seg-table-path 
                ./dseg.tsv --ref-region-label 8 --suvr-start 1800 --suvr-end 3600'          
"""

def main():
    """Configure petpal-pib-cli command."""
    help_text='Command line interface for running PiB processing.\n\nThis command runs the' \
    ' following steps:\n\n1. Crop out unnecessary voxels to speed up further processing' \
    '\n2. Motion Correct PET image with sliding windows\n3. Register PET image to the' \
    ' anatomical image given\n4. Sum PET image over provided time frame\n5. Compute ' \
    'SUVR Parametric Image with provided reference label\n6. Write roi tacs for optional' \
    ' further analysis'
    parser = argparse.ArgumentParser(prog='petpal-pib-proc',
                                     description=help_text,
                                     epilog=_PIB_EXAMPLE_, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--sub',required=True,help='Subject or participant identifier. Note that this should not include "sub-", but only the value, i.e."--sub 001"')
    parser.add_argument('--ses',required=True,help='Session identifier. Note that this should not include "ses-", but only the value, i.e. "--ses 01"')
    parser.add_argument('--pet-path',required=True,help='Path to 4D pet image')
    parser.add_argument('--anat-path',required=True,help='Path to anatomical image')
    parser.add_argument('--seg-path',required=True,help='Path to segmentation image')
    parser.add_argument('--seg-table-path',required=True,help='Path to segmentation table file (see /shared/dseg.tsv in petpal for an example)')
    parser.add_argument('--ref-region-label',required=True,type=int,help='Integer label which corresponds to the reference region in the segmentation image')
    parser.add_argument('--suvr-start',required=False,default=0,type=int,help='Time (in seconds) to start SUVR calculation. Defaults to 0 (beginning of scan)')
    parser.add_argument('--suvr-end',required=False,default=-1,type=int,help='Time (in seconds) to end SUVR calculation. Default to -1 (the end of the scan)')
    parser.add_argument('--bids-root',required=False,default='.',help='Path to root of bids directory. Defaults to current path (".")')
    
    args = parser.parse_args()

    sub_id = args.sub
    ses_id = args.ses
    pet_path = args.pet_path
    seg_path = args.seg_path
    anat_path = args.anat_path
    bids_dir = args.bids_root
    seg_table_path = args.seg_table_path
    suvr_start = args.suvr_start
    suvr_end = args.suvr_end
    ref_region_label = args.ref_region_label

    pib_pipeline = BIDS_Pipeline(sub_id=sub_id,
                                 ses_id=ses_id,
                                 pipeline_name='petpal-pib-cli',
                                 raw_anat_img_path=anat_path,
                                 segmentation_img_path=seg_path,
                                 bids_root_dir=bids_dir,
                                 segmentation_label_table_path=seg_table_path,
                                 raw_pet_img_path=pet_path)


    preproc_container = StepsContainer(name='preproc')

    # Configure steps for preproc container
    thresh_crop_step = ImageToImageStep.default_threshold_cropping(input_image_path=pib_pipeline.pet_path)
    moco_step = ImageToImageStep.default_windowed_moco()
    registration_step = ImageToImageStep.default_register_pet_to_t1(reference_image_path=pib_pipeline.anat_path,
                                                                    half_life=petpal.utils.constants.HALF_LIVES['c11'])
    write_tacs_step = TACsFromSegmentationStep.default_write_tacs_from_segmentation_rois(segmentation_image_path=pib_pipeline.seg_img,
                                                                                         segmentation_label_map_path=pib_pipeline.seg_table)
    wss_step = ImageToImageStep(name='weighted_series_sum',
                                function=petpal.utils.useful_functions.weighted_series_sum,
                                input_image_path='',
                                output_image_path='',
                                half_life=petpal.utils.constants.HALF_LIVES['c11'],
                                start_time=suvr_start,
                                end_time=suvr_end)

    # Add steps to preproc container
    preproc_container.add_step(step=thresh_crop_step)
    preproc_container.add_step(step=moco_step)
    preproc_container.add_step(step=registration_step)
    preproc_container.add_step(step=write_tacs_step)
    preproc_container.add_step(step=wss_step)

    kinetic_modeling_container = StepsContainer(name='km')

    # Configure steps for kinetic modeling container
    suvr_step = ImageToImageStep(name='suvr',
                                 function=petpal.preproc.standard_uptake_value.suvr,
                                 input_image_path='',
                                 output_image_path='',
                                 ref_region=ref_region_label,
                                 segmentation_image_path=seg_path)

    # Add steps to kinetic modeling container
    kinetic_modeling_container.add_step(step=suvr_step)

    pib_pipeline.add_container(step_container=preproc_container)
    pib_pipeline.add_container(step_container=kinetic_modeling_container)

    pib_pipeline.add_dependency(sending='thresh_crop', receiving='windowed_moco')
    pib_pipeline.add_dependency(sending='windowed_moco', receiving='register_pet_to_t1')
    pib_pipeline.add_dependency(sending='register_pet_to_t1', receiving='write_roi_tacs')
    pib_pipeline.add_dependency(sending='register_pet_to_t1', receiving='weighted_series_sum')
    pib_pipeline.add_dependency(sending='weighted_series_sum', receiving='suvr')

    pib_pipeline.update_dependencies()

    pib_pipeline()

if __name__ == "__main__":
    main()



"""
This module provides functions and a key class, :class:`GraphicalAnalysisParametricImage`, for 
graphical analysis and creation of parametric images of 4D-PET scan data. It heavily utilizes 
:mod:`numpy` for data manipulation and assumes the input as 4D PET images along with other required
inputs.

The :class:`GraphicalAnalysisParametricImage` class encapsulates the main functionality of the 
module, and encompasses methods for initializing data, running and saving analysis, calculating
various properties, and handling parametric image data.
"""

import os
import json
from collections.abc import Callable
from typing import Tuple, Union
import warnings
import ants
import numpy as np
import numba

from ..utils.dimension import gen_3d_img_from_timeseries

from .reference_tissue_models import fit_mrtm2_2003_to_tac,calc_bp_from_mrtm2_2003_fit
from .fit_tac_with_rtms import get_rtm_kwargs,get_rtm_method,get_rtm_output_size
from ..utils.time_activity_curve import TimeActivityCurve
from ..utils.dimension import (check_physical_space_for_ants_image_pair)
from .graphical_analysis import get_graphical_analysis_method, get_index_from_threshold
from ..input_function.blood_input import read_plasma_glucose_concentration
from ..utils.image_io import safe_copy_meta
from ..utils.time_activity_curve import safe_load_tac
from ..utils.dimension import gen_3d_img_from_timeseries

@numba.njit()
def apply_linearized_analysis_to_all_voxels(pTAC_times: np.ndarray,
                                            pTAC_vals: np.ndarray,
                                            tTAC_img: np.ndarray,
                                            t_thresh_in_mins: float,
                                            analysis_func: Callable) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generates parametric images for 4D-PET data using the provided analysis method.

    This function iterates over each voxel in the given `tTAC_img` and applies the provided
    `analysis_func` to compute analysis values. The `analysis_func` should be a numba.njit function
    for optimization and should be following a signature compatible with either of the following:
    patlak_analysis, logan_analysis, or alt_logan_analysis.

    Args:
        pTAC_times (np.ndarray): A 1D array representing the input TAC times in minutes.

        pTAC_vals (np.ndarray): A 1D array representing the input TAC values. This array should
                                be of the same length as `pTAC_times`.

        tTAC_img (np.ndarray): A 4D array representing the 3D PET image over time.
                               The shape of this array should be (x, y, z, time).

        t_thresh_in_mins (float): A float representing the threshold time in minutes.
                                  It is applied when calling the `analysis_func`.

        analysis_func (Callable): A numba.njit function to apply to each voxel for given PET data.
                                  It should take the following arguments:

                                    - input_tac_values: 1D numpy array for input TAC values
                                    - region_tac_values: 1D numpy array for regional TAC values
                                    - tac_times_in_minutes: 1D numpy array for TAC times in minutes
                                    - t_thresh_in_minutes: a float for threshold time in minutes

    Returns:
        Tuple[np.ndarray, np.ndarray]: A tuple of two 3D numpy arrays representing the calculated
            slope image and the intercept image, each of the same spatial dimensions as `tTAC_img`.

    """
    img_dims = tTAC_img.shape

    slope_img = np.zeros((img_dims[0], img_dims[1], img_dims[2]), float)
    intercept_img = np.zeros_like(slope_img)

    for i in range(0, img_dims[0], 1):
        for j in range(0, img_dims[1], 1):
            for k in range(0, img_dims[2], 1):
                analysis_vals = analysis_func(input_tac_values=pTAC_vals,
                                              region_tac_values=tTAC_img[i, j, k, :],
                                              tac_times_in_minutes=pTAC_times,
                                              t_thresh_in_minutes=t_thresh_in_mins)
                slope_img[i, j, k] = analysis_vals[0]
                intercept_img[i, j, k] = analysis_vals[1]

    return slope_img, intercept_img



@numba.njit()
def parametric_refregion_analysis(pTAC_times: np.ndarray,
                                  pTAC_vals: np.ndarray,
                                  tTAC_img: np.ndarray,
                                  t_thresh_in_mins: float,
                                  k2_prime: float,
                                  analysis_func: Callable) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generates parametric images for 4D-PET data using the provided analysis method. This function
    is intended only for use with reference region linear methods, such as Logan w/o arterial
    input.

    This function iterates over each voxel in the given `tTAC_img` and applies the provided
    `analysis_func` to compute analysis values. The `analysis_func` should be a numba.njit function
    for optimization and should be following a signature compatible with either of the following:
    patlak_analysis, logan_analysis, or alt_logan_analysis.

    Args:
        pTAC_times (np.ndarray): A 1D array representing the input TAC times in minutes.

        pTAC_vals (np.ndarray): A 1D array representing the input TAC values. This array should
                                be of the same length as `pTAC_times`.

        tTAC_img (np.ndarray): A 4D array representing the 3D PET image over time.
                               The shape of this array should be (x, y, z, time).

        t_thresh_in_mins (float): A float representing the threshold time in minutes.
                                  It is applied when calling the `analysis_func`.

        k2_prime (float): The population averaged reference region k2 value, passed to
                          `analysis_func`.
                                  
        analysis_func (Callable): A numba.njit function to apply to each voxel for given PET data.
                                  It should take the following arguments:

                                    - input_tac_values: 1D numpy array for input TAC values
                                    - region_tac_values: 1D numpy array for regional TAC values
                                    - tac_times_in_minutes: 1D numpy array for TAC times in minutes
                                    - t_thresh_in_minutes: a float for threshold time in minutes
                                    - k2_prime: the population averaged reference region k2 value
                    

    Returns:
        Tuple[np.ndarray, np.ndarray]: A tuple of two 3D numpy arrays representing the calculated
            slope image and the intercept image, each of the same spatial dimensions as `tTAC_img`.

    """
    img_dims = tTAC_img.shape

    slope_img = np.zeros((img_dims[0], img_dims[1], img_dims[2]), float)
    intercept_img = np.zeros_like(slope_img)

    for i in range(0, img_dims[0], 1):
        for j in range(0, img_dims[1], 1):
            for k in range(0, img_dims[2], 1):
                analysis_vals = analysis_func(input_tac_values=pTAC_vals,
                                              region_tac_values=tTAC_img[i, j, k, :],
                                              tac_times_in_minutes=pTAC_times,
                                              t_thresh_in_minutes=t_thresh_in_mins,
                                              k2_prime=k2_prime)
                slope_img[i, j, k] = analysis_vals[0]
                intercept_img[i, j, k] = analysis_vals[1]

    return slope_img, intercept_img


def generate_parametric_images_with_graphical_method(pTAC_times: np.ndarray,
                                                     pTAC_vals: np.ndarray,
                                                     tTAC_img: np.ndarray,
                                                     t_thresh_in_mins: float,
                                                     method_name: str,
                                                     **run_kwargs) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generates parametric images for 4D-PET data using a specified graphical analysis method.

    This function maps one of the predefined method names to the corresponding analysis function,
    and then generates parametric images by applying it to the given 4D-PET data using the
    `apply_linearized_analysis_to_all_voxels` function.

    Args:
        pTAC_times (np.ndarray): A 1D array representing the input TAC times in minutes.

        pTAC_vals (np.ndarray): A 1D array representing the input TAC values. This array should
                                be of the same length as `pTAC_times`.

        tTAC_img (np.ndarray): A 4D array representing the 3D PET image over time.
                               The shape of this array should be (x, y, z, time).

        t_thresh_in_mins (float): A float representing the threshold time in minutes.

        method_name (str): The analysis method's name to apply. Must be one of: 'patlak', 'logan',
            'alt_logan', or 'logan_ref'.

        run_kwargs: Keyword arguments with additional parameters for kinetic modeling. Currently
            only supports `k2_prime`.

    Returns:
        Tuple[np.ndarray, np.ndarray]: A tuple of two 3D numpy arrays representing the calculated
            slope image and the intercept image, each of the same spatial dimensions as `tTAC_img`.


    Raises:
        ValueError: If the `method_name` is not one of the following: 'patlak', 'logan',
            'alt_logan', 'logan_ref'.
    """
    if len(run_kwargs)>0:
        warnings.warn(f"Got the following run kwargs: {run_kwargs}. Kwargs other than 'k2_prime'"
                      "will be ignored.")
    analysis_func = get_graphical_analysis_method(method_name=method_name)
    if method_name!='logan_ref':
        slope_img, intercept_img = apply_linearized_analysis_to_all_voxels(pTAC_times=pTAC_times,
                                                                           pTAC_vals=pTAC_vals,
                                                                           tTAC_img=tTAC_img,
                                                                           t_thresh_in_mins=t_thresh_in_mins,
                                                                           analysis_func=analysis_func)
    else:
        slope_img, intercept_img = parametric_refregion_analysis(pTAC_times=pTAC_times,
                                                                 pTAC_vals=pTAC_vals,
                                                                 tTAC_img=tTAC_img,
                                                                 t_thresh_in_mins=t_thresh_in_mins,
                                                                 analysis_func=analysis_func,
                                                                 k2_prime=run_kwargs['k2_prime'])

    return slope_img, intercept_img


def apply_mrtm2_to_all_voxels(tac_times_in_minutes: np.ndarray,
                              tgt_image: np.ndarray,
                              ref_tac_vals: np.ndarray,
                              k2_prime: float,
                              t_thresh_in_mins: float,
                              mask_img: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generates parametric images for 4D-PET data using the MRTM2 reference tissue method.

    Args:
        tac_times_in_minutes (np.ndarray): A 1D array representing the reference TAC and PET frame
            times in minutes.
        tgt_image (np.ndarray): A 4D array representing the 3D PET image over time.
            The shape of this array should be (x, y, z, time).
        ref_tac_vals (np.ndarray): A 1D array representing the reference TAC values. This array
            should be of the same length as `tac_times_in_minutes`.
        k2_prime (float): A float representing the k2' value to be used for MRTM2 analysis. This
            is chosen based on the tracer or based on a regional MRTM1 analysis.
        t_thresh_in_mins (float): A float representing the threshold time past which MRTM
            parameters are calculated with a least squares fit.
        mask_img (np.ndarray): A 3D array representing the brain mask for `tgt_image`, where brain
            regions are labelled 1 and non-brain regions are labelled 0. This is made necessary in
            order to save time during computation. 

    Returns:
        bp_img (np.ndarray): A 3D array with computed BP values based on the MRTM2 parameter fit
            results. 
        simulation_img (np.ndarray): A 4D array with the same shape as `tgt_image` where each voxel
            is the best fit curve based on the solved parameters to the linear equation in MRTM2.
    """
    img_dims = tgt_image.shape

    bp_img = np.zeros((img_dims[0], img_dims[1], img_dims[2]), float)
    simulation_img = np.zeros_like(tgt_image)

    for i in range(0, img_dims[0], 1):
        for j in range(0, img_dims[1], 1):
            for k in range(0, img_dims[2], 1):
                if mask_img[i,j,k]>0.5:
                    analysis_vals = fit_mrtm2_2003_to_tac(tac_times_in_minutes=tac_times_in_minutes,
                                                          ref_tac_vals=ref_tac_vals,
                                                          tgt_tac_vals=tgt_image[i, j, k, :],
                                                          k2_prime=k2_prime,
                                                          t_thresh_in_mins=t_thresh_in_mins)
                    try:
                        bp_img[i, j, k] = calc_bp_from_mrtm2_2003_fit(analysis_vals[0])
                    except ValueError as exc:
                        print("Error in estimating BP from parameters, setting BP to NaN."
                              f"See: {exc}")
                        bp_img[i, j, k] = np.nan
                    simulation_img[i, j, k, :] = analysis_vals[1]

    return bp_img, simulation_img


def apply_rtm2_to_all_voxels(tac_times_in_minutes: np.ndarray,
                             tgt_image: np.ndarray,
                             ref_tac_vals: np.ndarray,
                             mask_img: np.ndarray,
                             method: str = 'mrtm2',
                             **analysis_kwargs) -> np.ndarray:
    """
    Generates parametric images for 4D-PET data using the SRTM2 reference tissue method.

    Args:
        tac_times_in_minutes (np.ndarray): A 1D array representing the reference TAC and PET frame
            times in minutes.
        tgt_image (np.ndarray): A 4D array representing the 3D PET image over time.
            The shape of this array should be (x, y, z, time).
        ref_tac_vals (np.ndarray): A 1D array representing the reference TAC values. This array
            should be of the same length as `tac_times_in_minutes`.
        mask_img (np.ndarray): A 3D array representing the brain mask for `tgt_image`, where brain
            regions are labelled 1 and non-brain regions are labelled 0. This is made necessary in
            order to save time during computation. 

    Returns:
        params_img (np.ndarray): A 4D array with RTM parameter fit results based on the supplied
            method.
    """
    bounds = False
    for kwarg in analysis_kwargs:
        if "_bounds" in kwarg:
            bounds = True
    analysis_func = get_rtm_method(method=method,bounds=bounds)
    img_dims = tgt_image.shape
    output_shape = get_rtm_output_size(method=method)
    params_img = np.zeros((img_dims[0], img_dims[1], img_dims[2], output_shape), float)

    for i in range(0, img_dims[0], 1):
        for j in range(0, img_dims[1], 1):
            for k in range(0, img_dims[2], 1):
                if mask_img[i,j,k]>0.5:
                    analysis_vals = analysis_func(tac_times_in_minutes=tac_times_in_minutes,
                                                  ref_tac_vals=ref_tac_vals,
                                                  tgt_tac_vals=tgt_image[i, j, k, :],
                                                  **analysis_kwargs)
                    params_img[i,j,k] = analysis_vals[0]

    return params_img


def generate_cmrglc_parametric_image_from_ki_image(input_ki_image_path: str,
                                                   output_image_path: str,
                                                   plasma_glucose_file_path: str,
                                                   glucose_rescaling_constant: float,
                                                   lumped_constant: float,
                                                   rescaling_const: float):
    r"""
    Generate and save a CMRglc image by rescaling a Patlak-Ki image.

    This function reads a Patlak-Ki image, rescales it using provided parameters (plasma glucose file,
    lumped constant, and a rescaling constant), and saves the resulting image as a CMRglc image.

    The final image will be `rescaling_constant * K_i * plasma_glucose / lumped_constant`.

    Args:
        input_ki_image_path (str): Path to the Patlak-Ki image file.
        output_image_path (str): Path to save the rescaled CMRglc image.
        plasma_glucose_file_path (str): File path to stored plasma glucose concentration.
            Assumed to be just one number in the file.
        glucose_rescaling_constant (float): Rescaling constant for the glucose concentration.
        lumped_constant (float): Lumped constant value used for rescaling.
        rescaling_const (float): Additional rescaling constant applied to the Patlak-Ki values.

    Returns:
        None
    """
    patlak_image = ants.image_read(filename=input_ki_image_path)
    plasma_glucose = read_plasma_glucose_concentration(file_path=plasma_glucose_file_path,
                                                       correction_scale=glucose_rescaling_constant)
    cmr_image = (plasma_glucose / lumped_constant) * patlak_image * rescaling_const
    ants.image_write(cmr_image, f"{output_image_path}")
    safe_copy_meta(input_image_path=input_ki_image_path, out_image_path=output_image_path)


class ReferenceTissueParametricImage:
    """
    Class for generating parametric images of 4D-PET images using reference tissue model (RTM)
    methods.

    Example:
        .. code-block:: python
            
            from petpal.kinetic_modeling import parametric_images
            
            rtm_parametric = ReferenceTissueParametricImage(reference_tac_path='/path/to/tac.tsv',
                                                            pet_image_path='/path/to/pet.nii.gz',
                                                            mask_image_path='/path/to/mask.nii.gz',
                                                            output_directory='/path/to/output,
                                                            output_filename_prefix='sub-001_mrtm2')
            rtm_parametric.run_parametric_analysis(method='mrtm2',
                                                   k2_prime=0.01,
                                                   t_thresh_in_mins=30)
            rtm_parametric.save_parametric_images()

    """
    def __init__(self,
                 reference_tac_path: str,
                 pet_image_path: str,
                 mask_image_path: str,
                 output_directory: str,
                 output_filename_prefix: str,
                 method: str='mrtm2'):
        """
        Initialize ReferenceTissueParametricImage with input values.

        Args:
            reference_tac_path (str): Path to the reference region TAC file.
            pet_image_path (str): Path to the 4D PET image on which kinetic analysis is performed.
            mask_image_path (str): Path to image that masks the brain in the same space as the PET
                image.
            output_directory (str): Path to folder where analysis is saved.
            output_filename_prefix (str): Prefix for output files saved after analysis.
            method (str): RTM method to run. Default 'mrtm2'.

        Raises:
            ValueError: When pet_image_path and mask_image_path are not in same physical space.
        """
        self.reference_tac = TimeActivityCurve.from_tsv(filename=reference_tac_path)
        self.pet_image = ants.image_read(pet_image_path)
        self.mask_image = ants.image_read(mask_image_path)

        if not check_physical_space_for_ants_image_pair(self.pet_image,self.mask_image):
            raise ValueError(f'Input image {pet_image_path} and mask {mask_image_path} not in'
                             'same physical space.')

        self.output_directory = output_directory
        self.output_filename_prefix = output_filename_prefix
        self.method = method
        self.analysis_props = self.init_analysis_props(method)
        self.fit_results = None, None


    def init_analysis_props(self, method: str) -> dict:
        r"""
        Initializes the analysis properties dict based on the specified RTM analysis method.

        Args:
            method (str): RTM analysis method. Must be one of 'srtm', 'frtm', 'mrtm-original',
                'mrtm' or 'mrtm2'.

        Returns:
            dict: A dictionary containing method-specific property keys and default values.

        Raises:
            ValueError: If input `method` is not one of the supported RTM methods.
        """
        common_props = {'MethodName': method.upper()}
        if method.startswith("mrtm"):
            props = {
                'BP': None,
                'k2Prime': None,
                'ThresholdTime': None,
                'Bounds': None,
                'StartFrameTime': None,
                'EndFrameTime' : None,
                'NumberOfPointsFit': None,
                'RawFits': None,
                **common_props
                }
        elif method.startswith("srtm") or method.startswith("frtm"):
            props = {
                'FitValues': None,
                'FitStdErr': None,
                **common_props
                }
        else:
            raise ValueError(f"Invalid method! Must be either 'srtm', 'frtm', 'srtm2', 'frtm2', "
                             f"'mrtm-original', 'mrtm' or 'mrtm2'. Got {method}.")
        return props


    def set_analysis_props(self,
                           props: dict,
                           bounds: Union[None, np.ndarray] = None,
                           k2_prime: float=None,
                           t_thresh_in_mins: float=None):
        """
        Set kwargs used for running parametric analysis.

        Args:
            rtm_kwargs (dict): Dictionary of kwargs fed into RTM analysis.
        """
        props['Bounds'] = bounds
        props['k2Prime'] = k2_prime
        props['ThresholdTime'] = t_thresh_in_mins


    def run_parametric_analysis(self,
                                bounds: Union[None, np.ndarray] = None,
                                k2_prime: float=None,
                                t_thresh_in_mins: float=None):
        """
        Run the analysis.

        Args:
            method (str): The method to be used in voxel-wise analysis. Currently only mrtm2 is
                implemented.
            bounds (Union[None, np.ndarray]): Bounds on fit parameters. See
                :py:func:`get_rtm_kwargs`. Default None.
            k2_prime (float): k2' value set for all voxel-wise analysis. Default None.
            t_thresh_in_mins (float): Threshold time after which kinetic parameters are fit.
                Default None.
        
        Returns:
            fit_results (np.ndarray, Tuple[np.ndarray, np.ndarray]): Kinetic parameters and
                simulated data returned as arrays. 
        """
        pet_np = self.pet_image.numpy()
        mask_np = self.mask_image.numpy()
        tac_times_in_minutes = self.reference_tac.times
        ref_tac_vals = self.reference_tac.activity
        rtm_method = get_rtm_method(self.method)
        analysis_kwargs = get_rtm_kwargs(method=rtm_method,
                                         bounds=bounds,
                                         k2_prime=k2_prime,
                                         t_thresh_in_mins=t_thresh_in_mins)

        fit_results = apply_rtm2_to_all_voxels(tac_times_in_minutes=tac_times_in_minutes,
                                               tgt_image=pet_np,
                                               ref_tac_vals=ref_tac_vals,
                                               mask_img=mask_np,
                                               method=self.method,
                                               **analysis_kwargs)
        self.fit_results = fit_results


    def save_parametric_images(self):
        """
        Save parametric images.
        """
        fit_arr = self.fit_results
        pet_img = self.pet_image
        if self.method=='mrtm2':
            bp_img_template = gen_3d_img_from_timeseries(input_img=pet_img)
            bp_arr = -(fit_arr[:,:,:,0]/fit_arr[:,:,:,1] + 1)
            fit_img = ants.from_numpy_like(data=bp_arr, image=bp_img_template)
        else:
            fit_img = ants.from_numpy(data=fit_arr,
                                    origin=pet_img.origin,
                                    spacing=pet_img.spacing,
                                    direction=pet_img.direction)

        try:
            if self.method=='mrtm2':
                fit_image_path = os.path.join(self.output_directory,
                                        f"{self.output_filename_prefix}_fit-mrtm2_bp.nii.gz")
            else:
                fit_image_path = os.path.join(self.output_directory,
                                        f"{self.output_filename_prefix}_desc-rtmfit_pet.nii.gz")
            ants.image_write(fit_img,fit_image_path)
        except IOError as exc:
            print("An IOError occurred while attempting to write the NIfTI image files.")
            raise exc from None


    def save_analysis_properties(self):
        """
        Saves the analysis properties to a JSON file in the output directory.

        This method involves saving a dictionary of analysis properties, which include file paths,
        analysis method, start and end frame times, threshold time, number of points fitted, and 
        various properties like the maximum, minimum, mean, and variance of slopes and intercepts
        found in the analysis. These analysis properties are written to a JSON file in the output
        directory with the name following the pattern
        `{output_filename_prefix}-analysis-props.json`.

        Args:
            None

        Returns:
            None

        Raises:
            IOError: An error occurred accessing the output_directory or while writing to the JSON
            file.

        See Also:
            * :func:`save_analysis_properties`
        """
        analysis_props_file = os.path.join(self.output_directory,
                                           f"{self.output_filename_prefix}_desc-"
                                           f"{self.analysis_props['MethodName']}_props.json")
        with open(analysis_props_file, 'w', encoding='utf-8') as f:
            json.dump(obj=self.analysis_props, fp=f, indent=4)


    def __call__(self,
                 bounds: np.ndarray=None,
                 t_thresh_in_mins: float=None,
                 k2_prime: float=None):
        self.run_parametric_analysis(bounds=bounds,
                                     t_thresh_in_mins=t_thresh_in_mins,
                                     k2_prime=k2_prime)
        self.set_analysis_props(props=self.analysis_props,
                                bounds=bounds,
                                k2_prime=k2_prime,
                                t_thresh_in_mins=t_thresh_in_mins)
        self.save_parametric_images()
        self.save_analysis_properties()

class GraphicalAnalysisParametricImage:
    """
    Class for generating parametric images of 4D-PET images using graphical analyses. It provides
    methods to run graphical analysis, calculate properties of the resulting images, and save the
    results using file paths.

    Attributes:
        input_tac_path (str): Absolute path to the input Time-Activity Curve (TAC) file.
        input_image_path (str): Absolute path to the 4D PET image file.
        output_directory (str): Absolute path to the output directory.
        output_filename_prefix (str): Prefix of the output file names.
        analysis_props (dict): Dictionary of properties of the graphical analysis.
        slope_image (np.ndarray): The slope image resulting from the graphical analysis,
            initialized to None.
        intercept_image (np.ndarray): The intercept image resulting from the graphical analysis,
            initialized to None.

    """

    def __init__(self,
                 input_tac_path: str,
                 input_image_path: str,
                 output_directory: str,
                 output_filename_prefix: str) -> None:
        """
        Initializes the GraphicalAnalysisParametricImage with the specified parameters.

        This method initializes necessary attributes for the GraphicalAnalysisParametricImage
        object with the provided arguments. It sets the absolute file paths for the input TAC, 4D
        PET image, and output directory, and initializes the analysis properties. Further, it
        initializes variables for the slope and intercept images.

        Args:
            input_tac_path (str): Path to the input Time-Activity Curve (TAC) file.
            input_image_path (str): Path to the 4D PET image file.
            output_directory (str): Path to the destination directory where output files will be
                saved.
            output_filename_prefix (str): Prefix to use for the names of the output files.

        Returns:
            None
        """
        self.input_tac_path = os.path.abspath(input_tac_path)
        self.input_image_path = os.path.abspath(input_image_path)
        self.pet_img = ants.image_read(filename=input_image_path)
        self.output_directory = os.path.abspath(output_directory)
        self.output_filename_prefix = output_filename_prefix
        self.analysis_props = self.init_analysis_props()
        self.slope_image: np.ndarray = None
        self.intercept_image: np.ndarray = None

    def init_analysis_props(self):
        """
        Initializes the analysis properties dictionary.

        The analysis properties dictionary contains properties derived from the analysis.
        It begins with certain known values, such as file paths, but most values are initialized
        to None and filled in later as the analysis is performed.

        Properties include:
            * ``FilePathPTAC`` (str): The path to the input Time-Activity Curve (TAC) file.
            * ``FilePathTTAC`` (str): The path to the 4D PET image file.
            * ``MethodName`` (str): The name of the graphical analysis method used, to be filled in later.
            * ``ImageDimensions`` (tuple): The dimensions of the images resulting from the analysis, to be filled in later.
            * ``StartFrameTime`` (float): The start time of the frame used in the analysis, filled in after the analysis.
            * ``EndFrameTime`` (float): The end time of the frame used in the analysis, filled in after the analysis.
            * ``ThresholdTime`` (float): The time threshold used in the analysis, filled in after the analysis.
            * ``RunKwargs`` (dict): Keyword arguments passed on to the analysis function.
            * ``NumberOfPointsFit`` (int): The number of points fitted in the analysis, filled in after the analysis.
            * ``SlopeMaximum`` (float): The maximum slope found in the analysis, filled in after the analysis.
            * ``SlopeMinimum`` (float): The minimum slope found in the analysis, filled in after the analysis.
            * ``SlopeMean`` (float): The mean of the slopes found in the analysis, filled in after the analysis.
            * ``SlopeVariance`` (float): The variance of the slopes found in the analysis, filled in after the analysis.
            * ``InterceptMaximum`` (float): The maximum intercept found in the analysis, filled in after the analysis.
            * ``InterceptMinimum`` (float): The minimum intercept found in the analysis, filled in after the analysis.
            * ``InterceptMean`` (float): The mean of the intercepts found in the analysis, filled in after the analysis.
            * ``InterceptVariance`` (float): The variance of the intercepts found in the analysis, filled in after the analysis.

        Returns:
            props (dict): The initialized properties dictionary.
        """
        props = {
            'FilePathPTAC': self.input_tac_path,
            'FilePathTTAC': self.input_image_path,
            'MethodName': None,
            'ImageDimensions': None,
            'StartFrameTime': None,
            'EndFrameTime': None,
            'ThresholdTime': None,
            'NumberOfPointsFit': None,
            'SlopeMaximum': None,
            'SlopeMinimum': None,
            'SlopeMean': None,
            'SlopeVariance': None,
            'InterceptMaximum': None,
            'InterceptMinimum': None,
            'InterceptMean': None,
            'InterceptVariance': None,
        }
        return props

    def run_analysis(self, method_name: str, t_thresh_in_mins: float, **run_kwargs):
        """
        Executes the complete analysis procedure.

        This method orchestrates the full analysis by orchestrating the calculation of parametric
        images, as well as compiling the properties related to the analysis. Both are determined
        based on the provided method name and the threshold time.

        Parameters:
            method_name (str): The name of the methodology adopted for the process.
            t_thresh_in_mins (float): The threshold time used through the analysis (in minutes).
            run_kwargs: Additional keyword arguments passed on to
                :func:`calculate_parametric_images` and :func:`calculate_analysis_properties`.

        See Also:
            * :func:`calculate_parametric_images`
            * :func:`calculate_analysis_properties`

        Returns:
            None

        """
        self.calculate_parametric_images(
            method_name=method_name, t_thresh_in_mins=t_thresh_in_mins, **run_kwargs)
        self.calculate_analysis_properties(
            method_name=method_name, t_thresh_in_mins=t_thresh_in_mins, **run_kwargs)

    def save_analysis(self):
        """
        Stores the results from an analysis routine.

        This method executes the storage of parametric images, as well as the properties related to
        the analysis. It assumes that the method 'run_analysis' is called before this method.

        Raises:
            RuntimeError: If the method 'run_analysis' is not called before this method.

        See Also:
            * :func:`save_parametric_images`
            * :func:`save_analysis_properties`

        Returns:
            None

        """
        if self.slope_image is None:
            raise RuntimeError(
                "'run_analysis' method must be called before 'save_analysis'.")
        self.save_parametric_images()
        self.save_analysis_properties()

    def calculate_analysis_properties(self,
                                      method_name: str,
                                      t_thresh_in_mins: float,
                                      **run_kwargs):
        """
        Performs a set of calculations to collate various analysis properties.

        This method orchestrates the calculation of properties related to both the parametric
        images and the fitting process. It does this by calling
        :meth:`calculate_parametric_images_properties` and :meth:`calculate_fit_properties`
        respectively.

        Args:
            method_name (str): The name of the method used for the fitting process.
            t_thresh_in_mins (float): The threshold time (in minutes) used for the fitting process.
            run_kwargs: Additional keyword arguments passed on to :func:`calculate_fit_properties`.

        See Also:
            * :meth:`calculate_parametric_images_properties`
            * :meth:`calculate_fit_properties`

        Returns:
            None. The results are stored within the instance's ``analysis_props`` variable.
        """
        self.calculate_parametric_images_properties()
        self.calculate_fit_properties(
            method_name=method_name, t_thresh_in_mins=t_thresh_in_mins, **run_kwargs)

    def calculate_fit_properties(self, method_name: str, t_thresh_in_mins: float, **run_kwargs):
        """
        Calculates and stores the properties related to the fitting process.

        This method calculates several properties related to the fitting process, including the
        threshold time, the name of the method used, the start and end frame time, and the number
        of points used in the fit. These values are stored in the instance's `analysis_props`
        variable.

        Args:
            method_name (str): The name of the methodology adopted for the fitting process.
            t_thresh_in_mins (float): The threshold time (in minutes) used in the fitting process.
            run_kwargs: Additional keyword arguments used in the analysis. These are saved to the
                analysis properties as individual properties.

        Note:
            This method relies on the :func:`safe_load_tac` function to load time-activity curve
            (TAC) data from the file at ``self.input_tac_path``, and the
            :func:`petpal.graphical_analysis.get_index_from_threshold` function to get the index
            from the threshold time.

        See also:
            * :func:`safe_load_tac`: Function to safely load TAC data from a file.
            * :func:`petpal.graphical_analysis.get_index_from_threshold`: Function to get the index
                from the threshold time.

        Returns:
            None. The results are stored within the instance's ``analysis_props`` variable.
        """
        self.analysis_props['ThresholdTime'] = t_thresh_in_mins
        self.analysis_props['MethodName'] = method_name

        for analysis_parameter_key, analysis_parameter_val in run_kwargs.items():
            self.analysis_props[analysis_parameter_key] = analysis_parameter_val

        p_tac_times, _ = safe_load_tac(filename=self.input_tac_path)
        t_thresh_index = get_index_from_threshold(times_in_minutes=p_tac_times,
                                                                     t_thresh_in_minutes=t_thresh_in_mins)
        self.analysis_props['StartFrameTime'] = p_tac_times[t_thresh_index]
        self.analysis_props['EndFrameTime'] = p_tac_times[-1]
        self.analysis_props['NumberOfPointsFit'] = len(
            p_tac_times[t_thresh_index:])

    def calculate_parametric_images_properties(self):
        """
        Initiates the calculation of properties for parametric images.

        This method triggers the calculation of statistical properties for slope and intercept
        images.
        Additionally, it captures the shape of the slope image as the 'ImageDimensions' and stores
        it in `analysis_props`.

        Note:
            You should ensure the `slope_image` attribute has been correctly set before calling
            this method. This means that `run_analysis` has already been called.

        See Also:
            calculate_slope_image_properties: Method to calculate various statistics for slope
                image.
            calculate_intercept_image_properties: Method to calculate various statistics for
                intercept image.

        Returns:
            None. The results are stored within the instance's `analysis_props` variable.
        """
        self.analysis_props['ImageDimensions'] = self.slope_image.shape
        self.calculate_slope_image_properties()
        self.calculate_intercept_image_properties()

    def calculate_slope_image_properties(self):
        """
        Calculates and stores statistical properties of the slope image.

        This method calculates the maximum, minimum, mean, and variance of
        the `slope_image` attribute, and stores these values in the `analysis_props` dictionary.

        The keys in `analysis_props` for these values are: `SlopeMaximum`, `SlopeMinimum`,
        `SlopeMean`, and `SlopeVariance`, respectively.

        Note:
            You should ensure the `slope_image` attribute has been correctly set before calling this
            method.

        No explicit return value. The results are stored within the instance's `analysis_props`
        variable.
        """
        self.analysis_props['SlopeMaximum'] = np.nanmax(self.slope_image)
        self.analysis_props['SlopeMinimum'] = np.nanmin(self.slope_image)
        self.analysis_props['SlopeMean'] = np.nanmean(self.slope_image)
        self.analysis_props['SlopeVariance'] = np.nanvar(self.slope_image)

    def calculate_intercept_image_properties(self):
        """
        Calculates and stores statistical properties of the intercept image.

        This method calculates the maximum, minimum, mean, and variance of
        the `intercept_image` attribute, and stores these values in the `analysis_props`
        dictionary.

        The keys in `analysis_props` for these values are: `InterceptMaximum`, `InterceptMinimum`,
        `InterceptMean`, and `InterceptVariance`, respectively.

        Note:
            You should ensure the `intercept_image` attribute has been correctly set before calling
            this method.

        No explicit return value. The results are stored within the instance's `analysis_props`
        variable.
        """
        self.analysis_props['InterceptMaximum'] = np.nanmax(self.intercept_image)
        self.analysis_props['InterceptMinimum'] = np.nanmin(self.intercept_image)
        self.analysis_props['InterceptMean'] = np.nanmean(self.intercept_image)
        self.analysis_props['InterceptVariance'] = np.nanvar(self.intercept_image)


    def calculate_parametric_images(self,
                                    method_name: str,
                                    t_thresh_in_mins: float,
                                    **run_kwargs):
        """
        Performs graphical analysis of PET parametric images and generates/updates the slope and
        intercept images.

        This method uses the given graphical analysis method and threshold to perform the analysis
        given the input Time Activity Curve (TAC) and 4D PET image, and updates the slope and 
        intercept images accordingly. PET images are loaded from the specified path. Then, the 
        parametric images are calculated using the specified graphical method and threshold time by
        explicitly analyzing each voxel in the 4D PET image.

        Args:
            method_name (str): The name of the graphical analysis method to be used.
            t_thresh_in_mins (float): The threshold time in minutes.
            run_kwargs: Additional keyword arguments passed on to
                :func:`generate_parametric_images_with_graphical_method`.

        Returns:
            None

        Raises:
            Exception: An error occurred during the graphical analysis. This could be due to an
            invalid method name or incorrect inputs to the method.

        See Also:
            * :func:`generate_parametric_images_with_graphical_method`
            * :func:`petpal.graphical_analysis.patlak_analysis`
            * :func:`petpal.graphical_analysis.logan_analysis`
            * :func:`petpal.graphical_analysis.alternative_logan_analysis`

        """
        p_tac_times, p_tac_vals = safe_load_tac(self.input_tac_path)
        self.slope_image, self.intercept_image = generate_parametric_images_with_graphical_method(
            pTAC_times=p_tac_times,
            pTAC_vals=p_tac_vals,
            tTAC_img=self.pet_img.numpy(),
            t_thresh_in_mins=t_thresh_in_mins, method_name=method_name,
            **run_kwargs)

    def __call__(self, method_name, t_thresh_in_mins, **run_kwargs):
        self.run_analysis(method_name=method_name, t_thresh_in_mins=t_thresh_in_mins, **run_kwargs)
        self.save_analysis()

    def save_parametric_images(self):
        """
        Saves the slope and intercept images as NIfTI files in the specified output directory.

        This method generates and saves two NIfTI files: one for the slope image and one for the
        intercept image. It uses the output directory and filename prefix provided during 
        instantiation of the class, along with the analysis method name, to generate a filename
        prefix for both images. The filenames follow the patterns 
        `{output_filename_prefix}-parametric-{method}-slope.nii.gz` and
        `{output_filename_prefix}-parametric-{method}-intercept.nii.gz` respectively. The affine
        transformation matrix for the new NIfTI images is derived from the original 4D PET image.

        Args:
            None

        Returns:
            None

        Raises:
            IOError: An error occurred accessing the output_directory or while writing to the NIfTI
            file.

        """
        file_name_prefix = os.path.join(self.output_directory,
                                        f"{self.output_filename_prefix}_desc-"
                                        f"{self.analysis_props['MethodName']}")
        template_img = gen_3d_img_from_timeseries(input_img=self.pet_img)
        try:
            tmp_slope_img = ants.from_numpy_like(data=self.slope_image, image=template_img)
            ants.image_write(tmp_slope_img, f"{file_name_prefix}_slope.nii.gz")

            tmp_intercept_img = ants.from_numpy_like(self.intercept_image, image=template_img)
            ants.image_write(tmp_intercept_img,f"{file_name_prefix}_intercept.nii.gz")

            safe_copy_meta(input_image_path=self.input_image_path,
                           out_image_path=f"{file_name_prefix}_slope.nii.gz")
            safe_copy_meta(input_image_path=self.input_image_path,
                           out_image_path=f"{file_name_prefix}_intercept.nii.gz")
        except IOError as e:
            print("An IOError occurred while attempting to write the NIfTI image files.")
            raise e from None

    def save_analysis_properties(self):
        """
        Saves the analysis properties to a JSON file in the output directory.

        This method involves saving a dictionary of analysis properties, which include file paths,
        analysis method, start and end frame times, threshold time, number of points fitted, and 
        various properties like the maximum, minimum, mean, and variance of slopes and intercepts
        found in the analysis. These analysis properties are written to a JSON file in the output
        directory with the name following the pattern
        `{output_filename_prefix}-analysis-props.json`.

        Args:
            None

        Returns:
            None

        Raises:
            IOError: An error occurred accessing the output_directory or while writing to the JSON
            file.

        See Also:
            * :func:`save_analysis_properties`
        """
        analysis_props_file = os.path.join(self.output_directory,
                                           f"{self.output_filename_prefix}_desc-"
                                           f"{self.analysis_props['MethodName']}_props.json")
        with open(analysis_props_file, 'w', encoding='utf-8') as f:
            json.dump(obj=self.analysis_props, fp=f, indent=4)



from . import graphical_plots
from . import image_visualization
from . import qc_plots
from . import tac_plots

def main():
    print("PETPAL - Pre-processing")


if __name__ == "__main__":
    main()
    


import tempfile
import os
import pandas as pd
import pytest
from petpal.utils.useful_functions import coerce_outpath_extension
from petpal.io.table import TableSaver

def test_coerce_to_csv():
    path_with_multiple_suffixes = 'file.suffix1.suffix2'
    path_expected = os.path.join(os.getcwd(),'file.csv')
    assert coerce_outpath_extension(path=path_with_multiple_suffixes, ext='.csv')==path_expected

def test_saver_saves_table():
    dataframe = pd.DataFrame(data={'a': 1},index=[0])
    saver = TableSaver()
    fd, tmp_path = tempfile.mkstemp(prefix="tmp_petpal_", dir='/tmp', suffix=".csv")
    os.close(fd)
    saver.save(df=dataframe, path=tmp_path)
    assert os.path.exists(tmp_path)
    with open(tmp_path) as f:
        assert f.read()==",a\n0,1\n"
    os.remove(tmp_path)

def test_saver_with_tab_separator():
    dataframe = pd.DataFrame(data={'a': 1},index=[0])
    saver = TableSaver()
    fd, tmp_path = tempfile.mkstemp(prefix="tmp_petpal_", dir='/tmp', suffix=".tsv")
    os.close(fd)
    saver.save(df=dataframe, path=tmp_path)
    assert os.path.exists(tmp_path)
    with open(tmp_path) as f:
        assert f.read()=="\ta\n0\t1\n"
    os.remove(tmp_path)

def test_saver_with_unaccepted_filetype():
    dataframe = pd.DataFrame(data={'a': 1},index=[0])
    saver = TableSaver()
    fd, tmp_path = tempfile.mkstemp(prefix="tmp_petpal_", dir='/tmp', suffix=".bad_extension")
    os.close(fd)
    with pytest.raises(Exception):
        saver.save(df=dataframe, path=tmp_path)


"""
This module contains a class to wrap the PETPVC package accessible via `PETPVC on GitHub <https://github.com/UCL/PETPVC>`_.
This package allows for numerous partial volume correction methods on PET data.

Note:
    This module is primarily a wrapper for the PETPVC package implemented via a Docker image
    hosted on `DockerHub <https://hub.docker.com/r/benthomas1984/petpvc>`_. The Docker image is Linux-based.

Requires:
    The module relies on the :mod:`docker` module.

"""

import os
import docker
from docker.errors import ImageNotFound, APIError, ContainerError
from typing import Union, Tuple


class PetPvc:
    """
    Handles operations for PET partial volume correction using a Docker container.

    This class manages the setup and execution of processes using the PETPVC package,
    accessible via `PETPVC on GitHub <https://github.com/UCL/PETPVC>`_ and a Docker image
    hosted on `DockerHub <https://hub.docker.com/r/benthomas1984/petpvc>`_. It manages
    image retrieval, input/output setup, and command execution.

    Attributes:
        client (docker.DockerClient): A Docker client connected to the local system.
        image_name (str): The Docker image name to use for PETPVC processes.

    Examples:
        Initialize the PETPVC handler and run a correction process:

        .. code-block:: python

            from petpvc_module import PetPvc
            petpvc_handler = PetPvc()
            petpvc_handler.run_petpvc(
                pet_4d_filepath="/path/to/input/pet_image.nii",
                output_filepath="/path/to/output/corrected_pet_image.nii",
                pvc_method="RBV",
                psf_dimensions=(6.0, 6.0, 6.0),
                mask_filepath="/path/to/input/brain_mask.nii",
                verbose=True)
                
    """

    def __init__(self):
        """
        Initializes the PetPvc instance and ensures the required Docker image is available.
        """
        self.client = docker.from_env()
        self.image_name = "benthomas1984/petpvc"
        self._pull_image_if_not_exists()

    def run_petpvc(self,
                   pet_4d_filepath: str,
                   output_filepath: str,
                   pvc_method: str,
                   psf_dimensions: Union[Tuple[float, float, float], float],
                   mask_filepath: str = None,
                   verbose: bool = False,
                   debug: bool = False) -> None:
        """
        Executes the PETPVC correction process within a Docker container.

        Args:
            pet_4d_filepath (str): The file path to the input 4D PET image.
            output_filepath (str): The file path where the output image should be saved.
            pvc_method (str): The partial volume correction method to apply.
            psf_dimensions (Union[Tuple[float, float, float], float]): The full-width half-max (FWHM) in mm along x, y, z axes.
            mask_filepath (str, optional): The file path to the mask image. Defaults to None.
            verbose (bool, optional): If True, prints the output from the Docker container. Defaults to False.
            debug (bool, optional): If True, adds the --debug flag to the command for detailed logs. Defaults to False.

        Prints:
            Outputs from the Docker container if verbose is True.

        Raises:
            docker.errors.ImageNotFound: If the Docker image is not available.
            docker.errors.APIError: If the Docker client encounters an API error.
        """
        pet_4d_filepath = os.path.abspath(pet_4d_filepath)
        output_filepath = os.path.abspath(output_filepath)
        docker_mask_input = None
        if mask_filepath:
            mask_filepath = os.path.abspath(mask_filepath)

        # Determine the common path
        common_path = os.path.commonpath(
            [pet_4d_filepath, output_filepath, mask_filepath]) if mask_filepath else os.path.commonpath(
            [pet_4d_filepath, output_filepath])
        if not common_path:
            common_path = os.path.dirname(pet_4d_filepath)
        if common_path == "/":
            docker_pet_input = f"/data/{pet_4d_filepath}"
            docker_output = f"/data/{output_filepath}"
            if mask_filepath:
                docker_mask_input = f"/data/{mask_filepath}"
        else:
            common_path = common_path.rstrip(os.sep)
            docker_pet_input = f"/data/{pet_4d_filepath.replace(common_path, '').lstrip(os.sep).replace(os.sep, '/')}"
            docker_output = f"/data/{output_filepath.replace(common_path, '').lstrip(os.sep).replace(os.sep, '/')}"
            if mask_filepath:
                docker_mask_input = f"/data/{mask_filepath.replace(common_path, '').lstrip(os.sep).replace(os.sep, '/')}"

        docker_volumes = {common_path: {'bind': '/data', 'mode': 'rw'}}

        command = f"petpvc --input {docker_pet_input} --output {docker_output} --pvc {pvc_method}"
        if mask_filepath:
            command += f" --mask {docker_mask_input}"
        if isinstance(psf_dimensions, tuple):
            command += f" -x {psf_dimensions[0]} -y {psf_dimensions[1]} -z {psf_dimensions[2]}"
        else:
            command += f" -x {psf_dimensions} -y {psf_dimensions} -z {psf_dimensions}"
        if debug:
            command += " --debug"

        print(f"Running Docker command: {command}")
        print(f"Volume bindings: {docker_volumes}")

        try:
            container = self.client.containers.run(self.image_name, command, volumes=docker_volumes, detach=False,
                                                   stream=True, auto_remove=False)
            if verbose:
                for line in container:
                    print(line.decode('utf-8').strip())
        except ContainerError as e:
            print(f"ContainerError: {e}")
            container_id = e.container.id
            logs = self.client.containers.get(container_id).logs(stream=True)
            for line in logs:
                print(line.decode('utf-8').strip())
            self.client.containers.get(container_id).remove()
            raise
        except APIError as e:
            print(f"APIError: {e}")
            raise

    def _pull_image_if_not_exists(self) -> None:
        """
        Checks if the Docker image is locally available and pulls it if not.
        """
        try:
            self.client.images.get(self.image_name)
            print(f"Image {self.image_name} is already available locally.")
        except ImageNotFound:
            print(f"Image {self.image_name} not found locally. Pulling from Docker Hub...")
            self.client.images.pull(self.image_name)
            print(f"Successfully pulled {self.image_name}.")
        except APIError as error:
            print(f"Failed to pull image due to API error: {error}")



import pytest
from petpal.preproc.symmetric_geometric_transfer_matrix import Sgtm

class DummyImage:
    def __init__(self, dimension):
        self.dimension = dimension

def make_sgtm_with_dimension(dim, sgtm_result=None):
    sgtm = object.__new__(Sgtm)
    sgtm.input_image = DummyImage(dimension=dim)
    sgtm.sgtm_result = sgtm_result
    return sgtm

def test_save_calls_save_results_3d_for_3d_image():
    sgtm = make_sgtm_with_dimension(3, sgtm_result=("labels", "vals", 1.0))
    called = {}
    def fake_save_results_3d(sgtm_result, out_tsv_path):
        called['args'] = (sgtm_result, out_tsv_path)
    sgtm.save_results_3d = fake_save_results_3d

    sgtm.save(output_path="out.tsv")

    assert 'args' in called
    assert called['args'][0] is sgtm.sgtm_result
    assert called['args'][1] == "out.tsv"

def test_save_calls_save_results_4d_tacs_when_one_tsv_per_region_true():
    sgtm = make_sgtm_with_dimension(4, sgtm_result="frame_results")
    called = {}
    def fake_save_results_4d_tacs(sgtm_result, out_tac_dir, out_tac_prefix):
        called['args'] = (sgtm_result, out_tac_dir, out_tac_prefix)
    sgtm.save_results_4d_tacs = fake_save_results_4d_tacs

    sgtm.save(output_path="/tmp/dir", out_tac_prefix="pref", one_tsv_per_region=True)

    assert 'args' in called
    assert called['args'][0] is sgtm.sgtm_result
    assert called['args'][1] == "/tmp/dir"
    assert called['args'][2] == "pref"

def test_save_calls_save_results_4d_multitacs_when_one_tsv_per_region_false():
    sgtm = make_sgtm_with_dimension(4, sgtm_result="frame_results")
    called = {}
    def fake_save_results_4d_multitacs(sgtm_result, out_tac_dir, out_tac_prefix):
        called['args'] = (sgtm_result, out_tac_dir, out_tac_prefix)
    sgtm.save_results_4d_multitacs = fake_save_results_4d_multitacs

    sgtm.save(output_path="/tmp/dir2", out_tac_prefix="pref2", one_tsv_per_region=False)

    assert 'args' in called
    assert called['args'][0] is sgtm.sgtm_result
    assert called['args'][1] == "/tmp/dir2"
    assert called['args'][2] == "pref2"


"""
This script provides a command-line interface (CLI) to calculate the image-derived input function (IDIF) using necktangle methods.
Users can choose between the single threshold method and the double threshold method.
The script supports the following functionalities:
- Single threshold method: Calculates the time-activity curve (TAC) from PET data using a single percentile threshold.
- Double threshold method: Computes the IDIF from a 4D PET necktangle matrix using a double threshold method.
Requires:
    The script relies on the :doc:`argparse <argparse:index>`, :doc:`numpy <numpy:index>`, and :doc:`nibabel <nibabel:index>` modules.
Example usage:
    Using the single threshold method:
        .. code-block:: bash
            python idif_necktangle_cli.py single --pet-4d-filepath /path/to/pet_4d.nii --carotid-necktangle-mask-3d-filepath /path/to/mask.nii --percentile 90
    Using the double threshold method:
        .. code-block:: bash
            python idif_necktangle_cli.py double --necktangle-filepath /path/to/necktangle.npy --percentile 90 --frame-start-times-filepath /path/to/start_times.npy --frame-duration-times-filepath /path/to/duration_times.npy
See Also:
    :mod:`pet_cli.idif_necktangle` - Module for calculating IDIF using necktangle methods.
"""

import argparse
import numpy as np
import nibabel as nib
from ..input_function.idif_necktangle import (
    single_threshold_idif_from_4d_pet_with_necktangle,
    double_threshold_idif_from_4d_pet_necktangle,
    get_frame_time_midpoints
)


def main():
    """
    Main function to handle command-line arguments and execute the appropriate IDIF calculation method.
    The script supports two methods:
    - Single threshold method
    - Double threshold method
    The user must provide the required file paths and parameters as command-line arguments.
    Raises:
        ValueError: If any input file paths are invalid or if required arguments are missing.
    """
    parser = argparse.ArgumentParser(description="Calculate IDIF using necktangle methods.")
    subparsers = parser.add_subparsers(dest="command", help="Sub-command help")

    # Single threshold sub-command
    single_parser = subparsers.add_parser("single", help="Use single threshold method")
    single_parser.add_argument("--pet-4d-filepath", type=str, required=True,
                               help="Path to the input 4D PET image file.")
    single_parser.add_argument("--carotid-necktangle-mask-3d-filepath", type=str, required=True,
                               help="Path to the carotid mask file.")
    single_parser.add_argument("--percentile", type=float, required=True, help="Percentile for threshold value.")
    single_parser.add_argument("--bolus_start_frame", type=int, default=3, help="Bolus start frame (default: 3).")
    single_parser.add_argument("--bolus_end_frame", type=int, default=7, help="Bolus end frame (default: 7).")

    # Double threshold sub-command
    double_parser = subparsers.add_parser("double", help="Use double threshold method")
    double_parser.add_argument("--necktangle-filepath", type=str, required=True,
                               help="Path to the 4D PET necktangle matrix file.")
    double_parser.add_argument("--percentile", type=float, required=True, help="Percentile for manual threshold value.")
    double_parser.add_argument("--frame-start-times-filepath", type=str, required=True,
                               help="Path to the frame start times file.")
    double_parser.add_argument("--frame-duration-times-filepath", type=str, required=True,
                               help="Path to the frame duration times file.")

    args = parser.parse_args()

    if args.command == "single":
        pet_4d_data = nib.load(args.pet_4d_filepath).get_fdata()
        carotid_necktangle_mask_3d_data = nib.load(args.carotid_necktangle_mask_3d_filepath).get_fdata()
        tac = single_threshold_idif_from_4d_pet_with_necktangle(
            pet_4d_data=pet_4d_data,
            carotid_necktangle_mask_3d_data=carotid_necktangle_mask_3d_data,
            percentile=args.percentile,
            bolus_start_frame=args.bolus_start_frame,
            bolus_end_frame=args.bolus_end_frame
        )
        np.savetxt("single_threshold_tac.csv", tac, delimiter=",")
        print("Single threshold IDIF calculation complete. TAC saved to single_threshold_tac.csv")

    elif args.command == "double":
        necktangle_matrix = np.load(args.necktangle_filepath)
        frame_start_times = np.load(args.frame_start_times_filepath)
        frame_duration_times = np.load(args.frame_duration_times_filepath)
        frame_midpoint_times = get_frame_time_midpoints(frame_start_times, frame_duration_times)
        tac = double_threshold_idif_from_4d_pet_necktangle(
            necktangle_matrix=necktangle_matrix,
            percentile=args.percentile,
            frame_midpoint_times=frame_midpoint_times
        )
        np.savetxt("double_threshold_tac.csv", tac, delimiter=",")
        print("Double threshold IDIF calculation complete. TAC saved to double_threshold_tac.csv")


if __name__ == "__main__":
    main()



"""
Provides methods to motion correct 4D PET data. Includes method
:meth:`determine_motion_target`, which produces a flexible target based on the
4D input data to optimize contrast when computing motion correction or
registration.
"""
from typing import Optional
from warnings import warn
import ants
import numpy as np
import pandas as pd
from scipy.spatial.transform import Rotation

from .motion_target import determine_motion_target
from ..utils.scan_timing import (ScanTimingInfo,
                                 get_window_index_pairs_from_durations)
from ..utils.useful_functions import (weighted_series_sum_over_window_indices,
                                      coerce_outpath_extension)
from ..utils.dimension import timeseries_from_img_list
from ..utils.image_io import get_half_life_from_nifti, safe_copy_meta
from ..io.table import TableSaver
from ..io.image import ImageLoader


class MotionCorrect:
    """Run windowed motion correction on an image and save the result.
    
    :ivar image_loader: :func:`~petpal.io.image.ImageLoader` instance or injectable replacement
    :ivar table_saver: :func:`~petpal.io.table.TableSaver` instance or injectable replacement
    :ivar input_img: (ants.ANTsImage) Dynamic PET image
    :ivar target_img: (ants.ANTsImage) Static target image
    :ivar scan_timing: :func:`~petpal.utils.scan_timing.ScanTimingInfo` Dynamic PET scan timing.
    :ivar half_life: (float) Half life of the PET tracer in seconds.
    :ivar: reg_kwargs: (dict) Keyword arguments passed on to :py:func:`~ants.registration`"""
    def __init__(self,
                 image_loader: Optional[ImageLoader] = None,
                 table_saver: Optional[TableSaver] = None):
        self.image_loader = image_loader or ImageLoader()
        self.table_saver = table_saver or TableSaver()
        self.input_img = None
        self.target_img = None
        self.scan_timing = None
        self.half_life = None
        self.reg_kwargs = self.default_reg_kwargs

    @property
    def default_reg_kwargs(self) -> dict:
        """Default registration arguments passed on to :py:func:`~ants.registration`."""
        reg_kwargs_default = {'aff_metric'               : 'mattes',
                              'write_composite_transform': True,
                              'interpolator'             : 'linear',
                              'type_of_transform'        : 'DenseRigid'}
        return reg_kwargs_default

    def set_reg_kwargs(self, **reg_kwargs):
        """Modify the registration arguments passed on to :py:func:`~ants.registration`."""
        self.reg_kwargs.update(**reg_kwargs)

    def set_input_scan_properties(self, input_image_path: str):
        """Load input image and get half life and scan timing. Set as MotionCorrect attributes.
        
        Args:
            input_image_path (str): Path to dynamic PET image."""
        self.input_img = self.image_loader.load(filename=input_image_path)
        self.half_life = get_half_life_from_nifti(image_path=input_image_path)
        self.scan_timing = ScanTimingInfo.from_nifti(image_path=input_image_path)

    def set_target_img(self, input_image_path: str, motion_target_option: str | tuple):
        """Get the motion target, load it as an image, and set as an attribute.
        
        Args:
            input_image_path (str): Path to dynamic PET image.
            motion_target_option (str | tuple): Option for motion target. See
                :meth:`~petpal.preproc.motion_target.determine_motion_target.` for details."""
        motion_target_path = determine_motion_target(motion_target_option=motion_target_option,
                                                     input_image_path=input_image_path)
        self.target_img = self.image_loader.load(filename=motion_target_path)

    def window_index_pairs(self, window_duration: float=300) -> np.ndarray:
        """The pair of indices corresponding to each window in the image.
        
        Args:
            window_duration (float): Scan will be divided into windows of this duration in
                seconds.
            
        Returns:
            window_index_pairs (np.ndarray): Array of start and end frame indices for each
                window.
        
        See also:
            - :meth:`~petpal.utils.scan_timing.get_window_index_pairs_from_durations`
        """
        return get_window_index_pairs_from_durations(frame_durations=self.scan_timing.duration,
                                                     window_duration=window_duration)

    def window_target_img(self, start_index: int, end_index: int) -> ants.ANTsImage:
        """Calculates the sum over frames in the target image within the provided time window.
        
        Args:
            start_index (int): Index for the frame that the window begins on.
            end_index (int): Index for the frame that the window ends on.
        
        Returns:
            window_img (ants.ANTsImage): Sum of frames in the input image between `start_index`
                and `end_index`."""
        return weighted_series_sum_over_window_indices(input_image_4d=self.input_img,
                                                        output_image_path=None,
                                                        window_start_id=start_index,
                                                        window_end_id=end_index,
                                                        half_life=self.half_life,
                                                        image_frame_info=self.scan_timing)

    @staticmethod
    def ants_xfm_to_rigid_pars(ants_xfm: ants.ANTsTransform):
        """Convert an ants transform object to six parameters (3 translation, 3 rotation) and the
        center reference point."""
        xfm_in = np.reshape(ants_xfm.parameters,(4,3))
        rot_matrix = xfm_in[:3,:]
        translate_matrix = xfm_in[3,:]

        scipy_rotation = Rotation.from_matrix(rot_matrix)
        rot_pars = -scipy_rotation.as_euler('xyz',degrees=True)

        xfm_out = list(rot_pars)+list(translate_matrix)+list(ants_xfm.fixed_parameters)
        return xfm_out

    def register_windows(self, window_duration: float=300) -> list[ants.ANTsTransform]:
        """Run motion correction on the input image to the target image.

        Creates "windows" by summing over frames with total length equal to `window_duration` and
        registering the window to the target image. Returns the calculated transforms for each
        frame.

        Args:
            window_duration (float): Duration of each window to sum over.

        Returns:
            window_xfm_stack (list[ants.ANTsTransform]): The transform to apply to each frame
                calculated based on the window the frame is in."""
        window_xfm_stack = []
        window_index_pairs = self.window_index_pairs(window_duration=window_duration)

        for _, (start_index, end_index) in enumerate(zip(*window_index_pairs)):
            window_target_img = self.window_target_img(start_index=start_index,
                                                       end_index=end_index)
            window_registration = ants.registration(fixed=self.target_img,
                                                    moving=window_target_img,
                                                    **self.reg_kwargs)
            window_xfm = ants.read_transform(window_registration['fwdtransforms'])
            for _ in range(start_index, end_index):
                window_xfm_stack.append(window_xfm)

        return window_xfm_stack

    def apply_motion_correction(self, frame_xfms: list[ants.ANTsTransform]) -> ants.ANTsImage:
        """Apply transforms to input image.
        
        Args:
            frame_xfms (list[ants.ANTsTransform]): Transforms for each frame in the input image.

        Returns:
            moco_img (ants.ANTsImage): Motion corrected dynamic PET image.
        """
        input_img_list = ants.ndimage_to_list(self.input_img)
        moco_img_stack = []

        for frame_index, frame_img in enumerate(input_img_list):
            frame_xfm = frame_xfms[frame_index]
            moco_frame_img = frame_xfm.apply_to_image(image=frame_img,
                                                      reference=self.target_img)
            moco_img_stack.append(moco_frame_img)

        moco_img = timeseries_from_img_list(moco_img_stack)
        return moco_img

    def save_xfm_parameters(self, frame_xfms: list[ants.ANTsTransform], filename: str):
        """Save frame transform parameters as a table.

        Args:
            frame_xfms (np.ndarray): Rigid transform parameters ordered as rotation, translation,
                centerpoint, then X, Y, Z axis, totalling 9 parameters for each frame.
            filename (str): Path to where table will be saved, including extension.

        Raises:
            ValueError: If transform type does not containt 'Rigid'. Saving transform parameters is
                currently only available for rigid transforms."""
        frame_xfm_pars = [self.ants_xfm_to_rigid_pars(ants_xfm=xfm) for xfm in frame_xfms]

        if 'Rigid' not in self.reg_kwargs['type_of_transform']:
            raise ValueError("Saving transform parameters is only available for rigid "
                             "registrations. Current transform type: "
                             f"{self.reg_kwargs['type_of_transform']}")
        xfm_columns = ['rot_x',
                       'rot_y',
                       'rot_z',
                       'tra_x',
                       'tra_y',
                       'tra_z',
                       'cen_x',
                       'cen_y',
                       'cen_z']
        xfms_df = pd.DataFrame(data=frame_xfm_pars,
                               columns=xfm_columns)
        xfms_df.index.name = 'frame'
        csv_filename = coerce_outpath_extension(path=filename, ext='.csv')
        self.table_saver.save(xfms_df,csv_filename)

    def __call__(self,
                 input_image_path: str,
                 output_image_path: str,
                 motion_target_option: str | tuple,
                 window_duration: float = 300,
                 transform_type: str = 'DenseRigid',
                 **reg_kwargs) -> ants.ANTsImage:
        """Motion correct a dynamic PET image.

        Divides image into segments of duration in seconds `window_duration` and register each frame
        to a target image, using the same transformation on for every frame in each window.

        Args:
            input_image_path (str): Path to dynamic PET image.
            output_image_path (str): Path to which motion corrected image is saved.
            motion_target_option (str | tuple): Path to motion target image, or specify time window
                such as (0,600) or preset option such as 'weighted_series_sum'. See
                :py:func:`~petpal.preproc.motion_target.determine_motion_target`.
            transform_type (str):  Type of transform used in ants.registration. See
                https://antspyx.readthedocs.io/en/latest/registration.html. Default DenseRigid.
            window_duration (float): Duration of each window in seconds. Default 300.
            reg_kwargs (keyword arguments): Keyword arguments to pass on to the registration
                function. See :py:func:`~ants.registration`.

        Returns:
            moco_img (ants.ANTsImage): Motion corrected dynamic PET image.
        """
        self.set_input_scan_properties(input_image_path=input_image_path)
        self.set_target_img(input_image_path=input_image_path,
                            motion_target_option=motion_target_option)

        self.set_reg_kwargs(type_of_transform=transform_type, **reg_kwargs)

        frame_xfms = self.register_windows(window_duration=window_duration)
        moco_img = self.apply_motion_correction(frame_xfms=frame_xfms)


        if 'Rigid' in self.reg_kwargs['type_of_transform']:
            self.save_xfm_parameters(frame_xfms=frame_xfms, filename=output_image_path)
        else:
            warn("Saving transform parameters is only available for rigid registrations. Current "
                 f" transform type: {self.reg_kwargs['type_of_transform']}.")

        ants.image_write(image=moco_img, filename=output_image_path)
        safe_copy_meta(input_image_path=input_image_path, out_image_path=output_image_path)

        return moco_img

def windowed_motion_corr_to_target(input_image_path: str,
                                   out_image_path: str | None,
                                   motion_target_option: str | tuple,
                                   window_duration: float,
                                   type_of_transform: str = 'QuickRigid',
                                   interpolator: str = 'linear',
                                   copy_metadata: bool = True,
                                   **kwargs):
    """
    Performs windowed motion correction (MoCo) to align frames of a 4D PET image to a given target image.
    We compute a combined image over the frames in a window, which is registered to the target image.
    Then, for each frame within the window, the same transformation is applied. This can be useful for
    initial frames with low counts. By setting a small-enough window size, later frames can still be
    individually registered to the target image.

    .. important::
        The motion-target will determine the space of the output image. If we provide a T1 image
        as the `motion_target_option`, the output image will be in T1-space.

    Note:
        This function is deprecated. Use :py:func:`~petpal.preproc.motion_corr.MotionCorrect`
        instead.

    Args:
        input_image_path (str): Path to the input 4D PET image file.
        out_image_path (str | None): Path to save the resulting motion-corrected image. If
            None, don't save image to disk.
        motion_target_option (str | tuple): Option to determine the motion target. This can
            be a path to a specific image file, a tuple of frame indices to generate a target, or
            specific options recognized by :func:`determine_motion_target`.
        window_duration (float): Window size in seconds for dividing the image into time sections.
        type_of_transform (str): Type of transformation to use in registration (default: 'QuickRigid').
        interpolator (str): Interpolation method for the transformation (default: 'linear').
        **kwargs: Additional arguments passed to :func:`ants.registration`.

    Returns:
        ants.core.ANTsImage: Motion-corrected 4D image.

    Workflow:
        1. Reads the input 4D image and splits it into individual frames.
        2. Computes index windows based on the specified window size (`window_duration`).
        3. Extracts necessary frame timing information and the tracer's half-life.
        4. For each window:
            - Calculates a weighted sum image for the window.
              See :func:`petpal.utils.useful_functions.weighted_series_sum_over_window_indecies`.
            - Performs registration of the weighted sum image to the target image.
            - Applies the obtained transformations to each frame within the window.
        5. Combines the transformed frames into a corrected 4D image.
        6. Saves the output image to the specified path, if provided.

    Note:
        If `out_image_path` is provided, the corrected 4D image will be saved to the specified path.
    """
    reg_kwargs_default = {'aff_metric'               : 'mattes',
                          'write_composite_transform': True,
                          'type_of_transform': type_of_transform,
                          'interpolator': interpolator}
    reg_kwargs = {**reg_kwargs_default, **kwargs}

    motion_corrector = MotionCorrect()
    moco_img = motion_corrector(input_image_path=input_image_path,
                                output_image_path=out_image_path,
                                motion_target_option=motion_target_option,
                                window_duration=window_duration,
                                copy_metadata=copy_metadata,
                                save_xfm=False
                                **reg_kwargs)
    return moco_img



"""
Module for generating quality control plots.
"""
import numpy as np
import pandas as pd
import seaborn as sns


def motion_plot(framewise_displacement: np.ndarray,
                output_plot: str=None):
    """
    Plots the quantity of motion estimated by :meth:`ants.motion_correction`.
    Takes the framewise displacement returned by ANTs, and plots this quantity
    against frame number.

    Args:
        framewise_displacement (np.ndarray): Total movement, or displacement,
            estimated between consecutive frames.
        output_plot (str): File to which plot is saved. If `None`, do not write
            plot to file. Default value `None`.
    
    Returns:
        movement_plot (sns.lineplot): Plot of total movement between frames.
    """
    movement_dataframe = pd.DataFrame(columns=['Frame','Framewise Displacement (mm)'])
    movement_dataframe['Frame'] = np.arange(len(framewise_displacement))
    movement_dataframe['Framewise Displacement (mm)'] = framewise_displacement
    movement_plot = sns.lineplot(data=movement_dataframe,x='Frame',y='Framewise Displacement (mm)')

    if output_plot is None:
        return movement_plot
    movement_plot.get_figure().savefig(output_plot)
    return movement_plot



"""This module contains all constants used throughout the library.

PET radionuclide half life source: code borrowed from DynamicPET
(https://github.com/bilgelm/dynamicpet/blob/main/src/dynamicpet/petbids/petbidsjson.py), derived
from TPC (turkupetcentre.net/petanalysis/decay.html). This source is from:
Table of Isotopes, Sixth edition, edited by C.M. Lederer, J.M. Hollander, I. Perlman. WILEY, 1967.
"""

HALF_LIVES = {
    "c11": 1224,
    "n13": 599,
    "o15": 123,
    "f18": 6588,
    "cu62": 582,
    "cu64": 45721.1,
    "ga68": 4080,
    "ge68": 23760000,
    "br76": 58700,
    "rb82": 75,
    "zr89": 282240,
    "i124": 360806.4,
}


CONVERT_kBq_to_mCi_ = 37000.0
r"""float: Convert kBq/ml to mCi/ml. 37000 kBq = 1 mCi.
"""


"""
This module contains utilities for handling `Brain Imaging Data Structure (BIDS) <https://bids.neuroimaging.io/>`_
datasets. It simplifies the creation and management of BIDS projects, offering functions for building project scaffolds,
handling file paths, and managing neuroimaging data files. Key features include scaffolding BIDS projects, caching
filepaths for efficient retrieval, and supporting various neuroimaging file types through integration with `nibabel`.
"""
import os
import pathlib

from bids_validator import BIDSValidator


def add_description_to_bids_path(filepath: str,
                                 description: str) -> str:
    r"""Create a copy of a BIDS filepath string with a description entity inserted before the suffix.

    Note:
        If a description 'desc-' entity already exists, it will be overwritten.

    See Also:
        `Brain Imaging Data Structure (BIDS) <https://bids.neuroimaging.io/>`_

    Args:
        filepath (str): BIDS-compliant filepath.
        description (str): 'desc' label to add to filename.

    Returns:
        string with 'desc' entity added (i.e. 'filename_desc-[description]_pet.nii.gz')
    """

    original_path = pathlib.Path(filepath)
    original_stem = original_path.stem
    split_stem = original_stem.split("_")
    split_stem_without_desc = [x for x in split_stem if not x.startswith('desc-')]
    split_stem_without_desc.insert(-1, f'desc-{description}')
    stem_with_desc = "_".join(split_stem_without_desc)
    new_path = str(original_path).replace(original_stem, stem_with_desc)

    return new_path

def validate_filepath_as_bids(filepath: str) -> bool:
    """
    Validate whether a given filepath conforms to the Brain Imaging Data Structure (BIDS) standard.

    Args:
        filepath (str): The path to the file to be validated.

    Returns:
        bool: True if the file conforms to the BIDS standard, False otherwise.

    """
    validator = BIDSValidator()
    return validator.is_bids(filepath)


def validate_directory_as_bids(project_path: str) -> bool:
    """
    Validate whether all files in a given directory and its subdirectories (excluding specified ones)
    conform to the Brain Imaging Data Structure (BIDS) standard.

    Args:
        project_path (str): The root directory of the project to validate.

    Returns:
        bool: True if all files in the directory conform to the BIDS standard, False if any do not.

    Raises:
        FileNotFoundError: If the provided project_path does not exist or is inaccessible.

    Notes:
        Excludes directories typically not needed for BIDS validation, such as 'code', 'derivatives',
        'sourcedata', '.git', and 'stimuli'. Also skips directories starting with 'sub-' to focus on top-level structure.
    """
    excluded_dirs = {'code', 'derivatives', 'sourcedata', '.git', 'stimuli'}
    all_passed = True
    failed_file_paths = []

    for root, dirs, files in os.walk(project_path, topdown=True):
        dirs[:] = [d for d in dirs if d not in excluded_dirs and not d.startswith('sub-')]
        for file in files:
            filepath = os.path.join(root, file)
            if not validate_filepath_as_bids(filepath):
                failed_file_paths.append(filepath)
                all_passed = False

    if failed_file_paths:
        print("Failed file paths:")
        for path in failed_file_paths:
            print(path)
    else:
        print("All files passed validation.")

    return all_passed


def parse_path_to_get_subject_and_session_id(path):
    """
    Parses a file path to extract subject and session IDs formatted according to BIDS standards.

    This function expects the file name in the path to contain segments starting with 'sub-' and 'ses-'.
    If these are not found, it returns default values indicating unknown IDs.

    Args:
        path (str): The file path to extract identifiers from.

    Returns:
        tuple: A tuple containing the subject ID and session ID.
    """
    filename = pathlib.Path(path).name
    if ('sub-' in filename) and ('ses-' in filename):
        sub_ses_ids = filename.split("_")[:2]
        sub_id = sub_ses_ids[0].split('sub-')[-1]
        ses_id = sub_ses_ids[1].split('ses-')[-1]
        return sub_id, ses_id
    else:
        return "XXXX", "XX"

def snake_to_camel_case(snake_str):
    """
    Converts a snake_case string to CamelCase.

    The function breaks the input string by underscores and capitalizes each segment to generate CamelCase.

    Args:
        snake_str (str): The snake_case string to convert.

    Returns:
        str: The converted CamelCase string.
    """
    return "".join(x.capitalize() for x in snake_str.lower().split("_"))

def gen_bids_like_filename(sub_id: str, ses_id: str, suffix: str= 'pet', ext: str= '.nii.gz', **extra_desc) -> str:
    """
    Generates a filename following BIDS convention including subject and session information.

    The function constructs filenames by appending additional descriptors and file suffix with extension.

    Args:
        sub_id (str): The subject identifier.
        ses_id (str): The session identifier.
        suffix (str, optional): The suffix indicating the data type. Defaults to 'pet'.
        ext (str, optional): The file extension. Defaults to '.nii.gz'.
        **extra_desc: Additional keyword arguments for any extra descriptors.

    Returns:
        str: A BIDS-like formatted filename.
    """
    sub_ses_pre = f'sub-{sub_id}_ses-{ses_id}'
    file_parts = [sub_ses_pre, ]
    for name, val in extra_desc.items():
        file_parts.append(f'{name}-{val}')
    file_parts.append(f'{suffix}{ext}')
    file_name = "_".join(file_parts)
    return file_name

def gen_bids_like_dir_path(sub_id: str, ses_id: str, modality: str='pet', sup_dir: str= '../') -> str:
    """
    Constructs a directory path following BIDS structure with subject and session subdirectories.

    Args:
        sub_id (str): The subject identifier.
        ses_id (str): The session identifier.
        modality (str, optional): Modality directory name. Defaults to 'pet'.
        sup_dir (str, optional): The parent directory path. Defaults to '../'.

    Returns:
        str: A BIDS-like directory path.
    """
    path_parts = [f'{sup_dir}', f'sub-{sub_id}', f'ses-{ses_id}', f'{modality}']
    return os.path.join(*path_parts)

def gen_bids_like_filepath(sub_id: str, ses_id: str, bids_dir:str ='../',
                           modality: str='pet', suffix:str='pet', ext='.nii.gz', **extra_desc) -> str:
    """
    Creates a full file path using BIDS-like conventions for both directory structure and filename.

    It combines directory and file generation to provide an organized output path.

    Args:
        sub_id (str): The subject identifier.
        ses_id (str): The session identifier.
        bids_dir (str, optional): Base directory for BIDS data. Defaults to '../'.
        modality (str, optional): The type of modality. Defaults to 'pet'.
        suffix (str, optional): Suffix indicating the type. Defaults to 'pet'.
        ext (str, optional): The file extension. Defaults to '.nii.gz'.
        **extra_desc: Additional keyword arguments for any extra descriptors.

    Returns:
        str: Full file path in BIDS-like structure.
        
    See Also:
        - :func:`gen_bids_like_filename`
        - :func:`gen_bids_like_dir_path`
        
    """
    filename = gen_bids_like_filename(sub_id=sub_id, ses_id=ses_id, suffix=suffix, ext=ext, **extra_desc)
    filedir  = gen_bids_like_dir_path(sub_id=sub_id, ses_id=ses_id, sup_dir=bids_dir, modality=modality)
    return os.path.join(filedir, filename)
    



"""
A collection of decorators to handle boilerplate code. Most decorators will
extend the functionality of functions that only work with objects or arrays.
The extensions allow for the flexibility of passing in image
paths or the image objects themselves, and allow the users to pass an
optional path for saving the output of the decorated function.
"""

import functools
import ants


def ANTsImageToANTsImage(func):
    """
    A decorator for functions that process an ANTs image and output another ANTs image.
    Assumes that the argument of the passed in function is an ANTs image.

    This decorator is designed to extend functions that take an ANTs image as input
    and output another ANTs image. It supports seamless handling of input images
    provided as either file paths (str) or `ants.core.ANTsImage` objects. The resulting
    processed image can optionally be saved to a specified file path.

    Args:
        func (Callable): The function to be decorated. It should accept an ANTs image as
            the first argument and return a processed ANTs image.

    Returns:
        Callable: A wrapper function that:
            - Reads the input image if a file path (str) is provided.
            - Passes an `ants.core.ANTsImage` object to the decorated function.
            - Saves the output image to the specified file path if `out_path` is provided.

    Example:

        .. code-block:: python

            import ants
            from petpal.utils.decorators import ANTsImageToANTsImage
            from petpal.preproc.segmentaion_tools import calc_vesselness_measure_image

            # Defining the decorated function
            @ANTsImageToANTsImage
            def step_calc_vesselness_measure_image(input_image: ants.core.ANTsImage,
                                                   sigma_min: float = 2.0,
                                                   sigma_max: float = 8.0,
                                                   alpha: float = 0.5,
                                                   beta: float = 0.5,
                                                   gamma: float = 5.0,
                                                   morph_open_radius: int = 1,
                                                   **hessian_func_kwargs):
                return calc_vesselness_measure_image(input_image=input_image,
                                                     sigma_min=sigma_min,
                                                     sigma_max=sigma_max,
                                                     alpha=alpha,
                                                     beta=beta,
                                                     gamma=gamma,
                                                     morph_open_radius=morph_open_radius,
                                                     **hessian_func_kwargs)

            # Conventional use of calc_vesselness_measure_image

            input_img = ants.image_read('/path/to/3d/img/.nii.gz')
            vess_img = calc_vesselness_measure_image(input_img) # Using all default values
            ants.image_write(vess_img, '/path/to/out/img/.nii.gz')


            # Using the decorated version
            ## Using paths as inputs
            vess_img = step_calc_vesselness_measure_image('/path/to/3d/img/.nii.gz',
                                                          '/path/to/out/img/.nii.gz')

            ### Not saving output image
            vess_img = step_calc_vesselness_measure_image('/path/to/3d/img/.nii.gz',
                                                          None)

            ## Using images as inputs
            input_img = ants.image_read('/path/to/3d/img/.nii.gz')
            vess_img = step_calc_vesselness_measure_image(input_img,
                                                          '/path/to/out/img/.nii.gz')

            ## Ignoring the return value to just save image
            step_calc_vesselness_measure_image('/path/to/3d/img/.nii.gz',
                                               '/path/to/out/img/.nii.gz')


    Raises:
        TypeError: If `in_img` is not a string or `ants.core.ANTsImage`.

    Notes:
        - If `in_img` is provided as a file path, the image is read using `ants.image_read`.
        - The output image is written to the desired path using `ants.image_write` if
          `out_path` is specified.
    """

    @functools.wraps(func)
    def wrapper(in_img:ants.core.ANTsImage | str,
                out_path: str,
                *args, **kwargs):
        if isinstance(in_img, str):
            in_image = ants.image_read(in_img)
        elif isinstance(in_img, ants.core.ANTsImage):
            in_image = in_img
        else:
            raise TypeError('in_img must be str or ants.core.ANTsImage')
        out_img = func(in_image, *args, **kwargs)
        if out_path is not None:
            ants.image_write(out_img, out_path)
        return out_img
    return wrapper
